class HttpRequest:
    def __init__(self, environ):
        from flask import Request

        self.r = Request(environ)

    @property
    def POST(self):  # noqa: N802
        return self.r.form

    @property
    def GET(self):  # noqa: N802
        return self.r.args

    @property
    def method(self):
        return self.r.method

    @property
    def META(self):  # noqa: N802
        return self.r.environ

    def is_ajax(self):
        return self.r.environ.get("HTTP_X_REQUESTED_WITH", "").lower() == "xmlhttprequest"



import inspect
import json
from pathlib import Path

from django.conf import settings
from django.http import (
    HttpResponse,
    HttpResponseRedirect,
)
from django.utils import autoreload
from django.views.decorators.csrf import csrf_exempt

from iommi import (
    Asset,
    Page,
    Part,
    Style,
    html,
    render_if_needed,
    render_part,
)
from iommi._web_compat import mark_safe
from iommi.base import get_wrapped_view, items
from iommi.struct import Struct
from iommi.style import (
    get_global_style,
    register_style,
)

orig_reload = getattr(autoreload, 'trigger_reload', None)


class Middleware:
    """
    The live edit middleware enables editing of the source code of views with as-you-type results, inside the web browser.

    Note: This middleware needs to go *first* in the middleware list.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_view(self, request, callback, callback_args, callback_kwargs):
        if should_edit(request):
            return live_edit_dispatch(request)(
                request=request, view=callback, args=callback_args, kwargs=callback_kwargs
            )


def live_edit_dispatch(request):
    return {
        '': live_edit_view,
        'stop': live_edit_view,
        'style_showcase': style_showcase,
        'style_editor__edit': style_editor__edit,
        'style_editor__new': style_editor__new,
        'style_editor': style_editor__select,
    }[request.GET['_iommi_live_edit']]


def should_edit(request):
    return settings.DEBUG and '_iommi_live_edit' in request.GET


def include_decorators(node):
    while node.parent.type == 'decorated':
        node = node.parent
    return node


def find_node(*, name, node, node_type):
    """
    node_type should be either funcdef or classdef
    """
    if node.type == node_type:
        if getattr(node, 'name', Struct(value=None)).value == name:
            return node
    for child_node in getattr(node, 'children', []):
        r = find_node(node=child_node, name=name, node_type=node_type)
        if r is not None:
            return include_decorators(r)
    return None


def find_view(view, ast_of_entire_file):
    if isinstance(view, Part):
        return find_node(name=type(view).__name__, node=ast_of_entire_file, node_type='classdef')
    else:
        return find_node(name=view.__name__, node=ast_of_entire_file, node_type='funcdef')


# This class exists just to provide a way to style the page
class LiveEditPage(Page):
    pass


def get_class_ast(cls):
    import parso

    filename = inspect.getsourcefile(cls)
    with open(filename) as f:
        entire_file = f.read()

    is_unix_line_endings = '\r\n' not in entire_file

    ast_of_entire_file = parso.parse(entire_file)
    return (
        find_node(name=cls.__name__, node=ast_of_entire_file, node_type='classdef'),
        is_unix_line_endings,
        ast_of_entire_file,
        filename,
    )


def create_response_for_view(new_view, request, args, kwargs, **_):
    if isinstance(new_view, type) and issubclass(new_view, Part):
        response = new_view().as_view()(request, *args, **kwargs)
    else:
        response = new_view(request, *args, **kwargs)
    return render_if_needed(request, response)


def get_ast(view):
    import parso

    # Read the old code
    try:
        # view is a function based view
        filename = view.__globals__['__file__']
    except AttributeError:
        # view is an iommi class
        from iommi.debug import filename_and_line_num_from_part

        filename, _ = filename_and_line_num_from_part(view)

    with open(filename) as f:
        entire_file = f.read()
        ast_of_entire_file = parso.parse(entire_file)

    is_unix_line_endings = '\r\n' not in entire_file

    ast_of_old_code = find_view(view, ast_of_entire_file)

    if ast_of_old_code is None:
        return get_class_ast(view.__class__)

    return ast_of_old_code, is_unix_line_endings, ast_of_entire_file, filename


def live_edit_post_handler(request, code, view, filename, create_response, write_new_code_to_disk, **params):
    try:
        final_result = dangerous_execute_code(
            code, request, view, create_response=create_response, args=params['args'], kwargs=params['kwargs']
        )

        if orig_reload is not None:
            # A little monkey patch dance to avoid one reload of the runserver when it's just us writing the code to disk
            # This only works in django 2.2+
            def restore_auto_reload(filename):
                from django.utils import autoreload

                print('Skipped reload')
                autoreload.trigger_reload = orig_reload

            autoreload.trigger_reload = restore_auto_reload

        write_new_code_to_disk(code=code, filename=filename, view=view, **params)

        return final_result
    except Exception as e:
        import traceback

        traceback.print_exc()
        error = str(e)
        if not error:
            error = str(e.__class__)
        return HttpResponse(json.dumps(dict(error=error)))


def write_new_code_to_disk_for_view(ast_of_entire_file, ast_of_old_code, code, filename, view, **_):
    import parso

    if isinstance(view, Part):
        ast_of_new_code = find_node(name=view.__class__.__name__, node=parso.parse(code), node_type='classdef')
    else:
        ast_of_new_code = find_node(name=view.__name__, node=parso.parse(code), node_type='funcdef')
    ast_of_old_code.children[:] = ast_of_new_code.children
    new_code = ast_of_entire_file.get_code()
    with open(filename, 'w') as f:
        f.write(new_code)


@csrf_exempt
def live_edit_view(request, view, args=None, kwargs=None):
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    try:
        import parso
    except ImportError:
        return HttpResponse('Live edit requires parso. Please `pip install parso`.')

    view = get_wrapped_view(view)
    # Read the old code
    try:
        # view is a function based view
        filename = view.__globals__['__file__']
    except AttributeError:
        # view is an iommi class
        from iommi.debug import filename_and_line_num_from_part

        filename, _ = filename_and_line_num_from_part(view)

    def build_params(entire_file, **_):
        ast_of_entire_file = parso.parse(entire_file)

        ast_of_old_code = find_view(view, ast_of_entire_file)
        assert ast_of_old_code is not None
        return dict(
            ast_of_entire_file=ast_of_entire_file,
            ast_of_old_code=ast_of_old_code,
        )

    flow_direction = request.GET.get('_iommi_live_edit_flow') or 'column'

    if request.GET.get('_iommi_live_edit') == 'stop':
        Path(filename).touch()
        return HttpResponseRedirect('.')
    assert flow_direction in ('column', 'row')

    return live_edit_view_impl(
        request,
        view=view,
        filename=filename,
        build_params=build_params,
        create_response=create_response_for_view,
        get_code=lambda ast_of_old_code, **_: ast_of_old_code.get_code(),
        write_new_code_to_disk=write_new_code_to_disk_for_view,
        flow_direction=flow_direction,
        args=args,
        kwargs=kwargs,
    )


def live_edit_view_impl(
    request,
    view,
    filename,
    build_params,
    get_code,
    create_response,
    write_new_code_to_disk,
    flow_direction,
    args,
    kwargs,
):
    with open(filename) as f:
        entire_file = f.read()

    is_unix_line_endings = '\r\n' not in entire_file

    params = {
        'entire_file': entire_file,
        'is_unix_line_endings': is_unix_line_endings,
        'view': view,
        'filename': filename,
        'request': request,
        'create_response': create_response,
        'write_new_code_to_disk': write_new_code_to_disk,
        'args': args,
        'kwargs': kwargs,
    }
    params = {
        **params,
        **build_params(**params),
    }

    if request.method == 'POST':
        code = request.POST['data'].replace('\t', '    ')
        if is_unix_line_endings:
            code = code.replace('\r\n', '\n')
        params['code'] = code
        return live_edit_post_handler(**params)

    code = get_code(**params)

    return LiveEditPage(
        title='iommi live edit',
        h_tag__include=False,
        assets__code_editor=Asset.js(
            attrs=dict(
                src='https://cdnjs.cloudflare.com/ajax/libs/ace/1.4.12/ace.js',
                integrity='sha512-GZ1RIgZaSc8rnco/8CXfRdCpDxRCphenIiZ2ztLy3XQfCbQUSCuk8IudvNHxkRA3oUg6q0qejgN/qqyG1duv5Q==',
                crossorigin='anonymous',
            ),
            after=-1,
        ),
        assets__live_edit_page_custom=Asset(
            tag='style',
            text='''
                .container {
                    padding: 0 !important;
                    margin: 0 !important;
                    max-width: 100%;
                }

                html,
                body {
                    height: 100%;
                    margin: 0;
                }

                .container {
                    display: flex;
                    flex-flow: <<flow_direction>>;
                    height: 100%;
                }

                .container iframe {
                    flex: 1 1 auto;
                    border: 0;
                }

                .container #editor_and_error {
                    flex: 2 1 auto;
                }

                #editor {
                    height: 90% !important;
                }

                #error {
                    height: 10% !important;
                }
            '''.replace(
                '<<flow_direction>>', flow_direction
            ),
        ),
        iommi_style='bootstrap',
        parts__result=html.iframe(attrs__id='result'),
        parts__editor_and_error=html.div(
            attrs__id='editor_and_error',
            children=dict(
                editor=html.div(
                    code,
                    attrs__id='editor',
                ),
                error=html.div(attrs__id='error'),
            ),
        ),
        parts__script=html.script(
            mark_safe(
                # language=javascript
                '''
                    var editor = ace.edit("editor");
                    editor.setTheme("ace/theme/cobalt");
                    editor.session.setMode("ace/mode/python");
                    editor.setShowPrintMargin(false);

                    async function update() {
                        if (window.iommi_live_edit_updating) {
                            return;
                        }

                        window.iommi_live_edit_updating = true;
                        window.iommi_live_edit_needs_update = false;

                        let form_data = new FormData();
                        form_data.append('data', editor.getValue());

                        let response = await fetch('', {
                            method: 'POST',
                            body: form_data
                        });
                        let foo = await response.json();
                        if (foo.page) {
                            // TODO: get scroll position and restore it
                            document.getElementById('result').srcdoc = foo.page;
                        }
                        document.getElementById('error').innerText = foo.error || '';
                        window.iommi_live_edit_updating = false;
                    }

                    setInterval(() => {
                        if (!window.iommi_live_edit_needs_update) {
                            return;
                        }
                        update();
                    }, 200)

                    window.iommi_live_edit_updating = false;
                    window.iommi_live_edit_needs_update = true;

                    editor.session.on('change', () => {window.iommi_live_edit_needs_update = true});
                    editor.setFontSize(14);
                    editor.session.setUseWrapMode(true);
                '''
            )
        ),
    )


def dangerous_execute_code(code, request, view, args, kwargs, create_response=create_response_for_view):
    local_variables = {}
    if isinstance(view, Part):
        from importlib import import_module

        module = import_module(view.__module__)
        exec(code, module.__dict__, local_variables)
    elif isinstance(view, Style):
        frame = view._instantiated_at_frame.f_back
        exec(code, frame.f_globals, local_variables)
    else:
        exec(code, view.__globals__, local_variables)
    request.method = 'GET'
    request.GET = request.GET.copy()
    request.GET['_iommi_disable_debug_panel'] = ''
    new_view = list(local_variables.values())[0]
    response = create_response(new_view, request=request, args=args, kwargs=kwargs, local_variables=local_variables)
    final_result = HttpResponse(json.dumps(dict(page=response.content.decode())))
    return final_result


def style_showcase(request, style=None, **_):
    from django.contrib.auth.models import User

    if style is None:
        from iommi.style import DEFAULT_STYLE

        style = getattr(settings, 'IOMMI_DEFAULT_STYLE', DEFAULT_STYLE)

    class DummyRow:
        def __init__(self, idx):
            self.idx = idx

        def __getattr__(self, attr):
            _, _, shortcut = attr.partition('column_of_type_')
            s = f'{shortcut} #{self.idx}'
            if shortcut == 'link' or attr == 'link':

                class Link:
                    def get_absolute_url(self):
                        return '#'

                    def __str__(self):
                        return 'title'

                return Link()
            if shortcut == 'number':
                return f'{self.idx}'
            return s

        @staticmethod
        def get_absolute_url():
            return '#'

    from iommi import (
        Action,
        Column,
        Field,
        Form,
        Header,
        Menu,
        MenuItem,
        Table,
    )

    return Page(
        iommi_style=style,
        parts=dict(
            title=Header('Style showcase'),
            menu__children=dict(
                menu_title=html.h2('Menu'),
                menu=Menu(
                    sub_menu=dict(
                        active=MenuItem(url=request.get_full_path()),  # full path do make this item active
                        inactive=MenuItem(),
                    )
                ),
            ),
            form=Form(
                title='Form',
                fields=dict(
                    text=Field.text(initial='initial'),
                    boolean=Field.boolean(),
                    boolean_selected=Field.boolean(initial=True),
                    radio=Field.radio(choices=['a', 'b', 'c'], initial='b'),
                ),
                actions__submit__post_handler=lambda **_: None,
                actions__secondary=Action.button(),
                actions__delete=Action.delete(display_name='Delete'),
                actions__icon=Action.icon('trash', display_name='Icon', attrs__href='#'),
            ),
            table=Table(
                title='Table',
                model=User,
                columns={
                    t.__name__: dict(call_target=t, display_name=t.__name__)
                    for t in [
                        Column.select,
                        Column.edit,
                        Column.delete,
                        Column.boolean,
                        Column.text,
                        Column.number,
                        Column.link,
                        Column.icon,
                    ]
                },
                columns__text__filter__include=True,
                columns__number__filter__include=True,
                rows=[DummyRow(i) for i in range(10)],
                page_size=2,
            ),
        ),
    )


def style_editor__select(**_):
    from iommi import Action, Field, Form
    from iommi.style import _styles

    return Form(
        title='Select style to edit',
        fields__name=Field.choice(choices=[k for k, v in items(_styles) if not v.internal]),
        actions__edit=Action.primary(
            display_name='Edit',
            post_handler=lambda form, **_: HttpResponseRedirect(
                f'?_iommi_live_edit=style_editor__edit&name={form.fields.name.value}'
            )
            if form.is_valid()
            else None,
        ),
        actions__new_style__attrs__href='?_iommi_live_edit=style_editor__new',
    )


@csrf_exempt
def style_editor__edit(request, **_):
    name = request.GET.get('name', request.POST.get('name'))
    assert name is not None
    style = get_global_style(name)
    filename = style._instantiated_at_frame.f_back.f_code.co_filename

    def create_response(view, local_variables, **_):
        view._instantiated_at_frame = (
            style._instantiated_at_frame
        )  # otherwise we end up at <string> which is not what we want
        style_editor_name = 'style_editor'
        with register_style(style_editor_name, style=local_variables[name], allow_overwrite=True):
            return render_part(
                request,
                style_showcase(request, style=style_editor_name),
            )

    def write_new_code_to_disk(code, **_):
        with open(filename, 'w') as f:
            f.write(code)

    return live_edit_view_impl(
        request,
        view=style,  # TODO: rename view.. "subject"?
        filename=filename,
        build_params=lambda **_: {},
        get_code=lambda entire_file, **_: entire_file,
        create_response=create_response,
        write_new_code_to_disk=write_new_code_to_disk,
        flow_direction='row',
        args=[],
        kwargs={},
    )


def style_editor__new(**_):
    from iommi import (
        Field,
        Form,
        Header,
    )

    def new_style(form, **_):
        if not form.is_valid():
            return

        base, _, new = form.fields.module.value.rpartition('.')
        module = __import__(base, fromlist=['_silly_importlib'])
        target_filename = Path(module.__file__).parent / (new + '.py')
        if target_filename.exists():
            form.add_error(f'File {target_filename} already exists')

        with open(target_filename, 'w') as f:
            f.write(
                f'''
from iommi.style import Style
from iommi.style_base import base
from iommi.asset import Asset

{new} = Style(
    base,
)
'''
            )
            return Page(
                parts=dict(
                    title=Header('Style created... now what?'),
                    message=html.p(
                        f'''
                            The style file was written to {target_filename}.
                            Now you need to register this style in order to edit it. This is
                            typically done by adding `register_style('{new}')` into `on_ready`
                            of your `AppConfig`.
                        '''
                    ),
                    message2=html.p("When you've done that, you can proceed to "),
                    edit=html.a('edit it', attrs__href=f'?_iommi_live_edit=style_editor__edit&name={new}'),
                ),
            )

    return Form(
        title='Create new style',
        fields=dict(
            module=Field(),  # TODO: can we guess this in a smart way? maybe look at settings.__module__?
        ),
        actions__submit__post_handler=new_style,
    )



import datetime
import operator
from functools import reduce
from typing import (
    Type,
    Union,
)

from django.conf import settings
from django.core.exceptions import (
    MultipleObjectsReturned,
    ObjectDoesNotExist,
)
from django.db.models import (
    F,
    Model,
    Q,
    QuerySet,
)
from django.utils import timezone
from django.utils.translation import (
    gettext,
    pgettext,
)
from pyparsing import (
    Char,
    Forward,
    Group,
    Keyword,
    ParseException,
    ParseResults,
    QuotedString,
    Word,
    ZeroOrMore,
    alphanums,
    alphas,
    oneOf,
    quotedString,
)

from iommi._web_compat import (
    Template,
    ValidationError,
    render_template,
)
from iommi.action import (
    Action,
)
from iommi.base import (
    MISSING,
    NOT_BOUND_MESSAGE,
    items,
    keys,
    model_and_rows,
    values,
)
from iommi.declarative import declarative
from iommi.declarative.dispatch import dispatch
from iommi.declarative.namespace import (
    EMPTY,
    Namespace,
    getattr_path,
    setdefaults_path,
)
from iommi.declarative.with_meta import with_meta
from iommi.endpoint import path_join
from iommi.evaluate import (
    evaluate,
    evaluate_strict,
)
from iommi.form import (
    Form,
    bool_parse,
    boolean_tristate__parse,
    date_parse,
    datetime_parse,
    float_parse,
    int_parse,
    time_parse,
)
from iommi.fragment import (
    Fragment,
)
from iommi.from_model import (
    AutoConfig,
    NoRegisteredSearchFieldException,
    create_members_from_model,
    get_search_fields,
    member_from_model,
)
from iommi.member import (
    bind_member,
    bind_members,
    refine_done_members,
)
from iommi.part import (
    Part,
    request_data,
)
from iommi.refinable import (
    EvaluatedRefinable,
    Prio,
    Refinable,
    RefinableMembers,
    SpecialEvaluatedRefinable,
    refinable,
)
from iommi.shortcut import (
    Shortcut,
    with_defaults,
)
from iommi.struct import Struct


class QueryException(Exception):
    pass


PRECEDENCE = {
    'and': 3,  # pragma: no mutate
    'or': 2,  # pragma: no mutate
}
assert PRECEDENCE['and'] > PRECEDENCE['or']  # pragma: no mutate

Q_OPERATOR_BY_QUERY_OPERATOR = {
    '>': 'gt',
    '=>': 'gte',
    '>=': 'gte',
    '<': 'lt',
    '<=': 'lte',
    '=<': 'lte',
    '=': 'iexact',
    ':': 'icontains',
}

FREETEXT_SEARCH_NAME = 'freetext_search'

_filter_factory_by_django_field_type = {}


def register_filter_factory(django_field_class, *, shortcut_name=MISSING, factory=MISSING, **kwargs):
    assert shortcut_name is not MISSING or factory is not MISSING
    if factory is MISSING:
        factory = Shortcut(call_target__attribute=shortcut_name, **kwargs)

    _filter_factory_by_django_field_type[django_field_class] = factory


def to_string_surrounded_by_quote(v):
    str_v = '%s' % v
    return '"%s"' % str_v.replace('"', '\\"')


def value_to_str_for_query(filter, v):
    if isinstance(v, bool):
        return {True: '1', False: '0'}.get(v)
    if type(v) in (int, float):
        return str(v)
    if isinstance(v, datetime.datetime) and timezone.is_aware(v):
        v = timezone.make_naive(v, timezone=datetime.timezone.utc)
    if isinstance(v, Model):
        model = type(v)
        search_field = filter.search_fields[-1]
        try:
            v = getattr_path(v, search_field)
        except AttributeError:
            raise NoRegisteredSearchFieldException(
                f'{model.__name__} has no attribute {search_field}. Please register search fields with register_search_fields or specify search_fields.'
            )
    return to_string_surrounded_by_quote(v)


def build_query_expression(*, filter, value):
    if isinstance(value, Model):
        return f'{filter.query_name}.pk={value.pk}'

    return f'{filter.query_name}{filter.query_operator_for_field}{value_to_str_for_query(filter, value)}'


def case_sensitive_query_operator_to_q_operator(op):
    return {'=': 'exact', ':': 'contains'}.get(op) or Q_OPERATOR_BY_QUERY_OPERATOR[op]


def boolean__query_operator_to_q_operator(op):
    if op != '=':
        raise QueryException(f'Invalid operator "{op}" for boolean filter. The only valid operator is "=".')
    return 'exact'


def choice_queryset_value_to_q(filter, op, value_string_or_f, **_):
    if op != '=':
        raise QueryException(f'Invalid operator "{op}" for filter "{filter._name}"')
    if filter.attr is None:
        return Q()
    if isinstance(value_string_or_f, str) and value_string_or_f.lower() == 'null':
        return Q(**{filter.attr: None})
    try:
        instance = None
        for search_field in filter.search_fields:
            try:
                instance = filter.choices.get(**{search_field: str(value_string_or_f)})
                break
            except ObjectDoesNotExist:
                pass
    except MultipleObjectsReturned:
        raise QueryException(f'Found more than one object for name "{value_string_or_f}"')
    if instance is None:
        return None
    return Q(**{filter.attr + '__pk': instance.pk})


choice_queryset_value_to_q.iommi_needs_attr = True


def default_filter__is_valid_filter(name, filter, **_):
    return (
        filter.attr or filter.value_to_q,
        f"{name} cannot be a part of a query, it has no attr or value_to_q so we don't know what to search for. If you want to include it anyway you can define the callback is_valid_filter which should return a boolean and a string with an error message if the boolean is False. The simplest way to do that would be is_valid_filter=lambda **_: (True, '') (filter__is_valid_filter=lambda **_: (True, '') for a Column)",
    )


def choice_queryset__is_valid_filter(name, filter, **_):
    return (
        filter.attr,
        f"{name} cannot be a part of a query, it has no attr so we don't know what to search for. If you want to include it anyway you can define the callback is_valid_filter which should return a boolean and a string with an error message if the boolean is False. The simplest way to do that would be is_valid_filter=lambda **_: (True, '') (filter__is_valid_filter=lambda **_: (True, '') for a Column)",
    )


@with_meta
class Filter(Part):
    """
    Class that describes a filter that you can search for.

    See :doc:`Query` for more complete examples.
    """

    attr = EvaluatedRefinable()
    field: Namespace = Refinable()
    query_operator_for_field: str = EvaluatedRefinable()
    freetext = EvaluatedRefinable()
    model: Type[Model] = SpecialEvaluatedRefinable()
    model_field = Refinable()
    model_field_name = Refinable()
    choices = EvaluatedRefinable()
    search_fields = Refinable()
    unary = Refinable()
    is_valid_filter = Refinable()
    query_name = EvaluatedRefinable()
    pk_lookup_to_q = Refinable()

    @with_defaults(
        query_operator_for_field='=',
        attr=MISSING,
        search_fields=MISSING,
        field__required=False,
        is_valid_filter=default_filter__is_valid_filter,
        query_name=lambda filter, **_: filter.iommi_name(),
    )
    def __init__(self, **kwargs):
        """
        Parameters with the prefix `field__` will be passed along downstream to the `Field` instance if applicable. This can be used to tweak the basic style interface.

        :param field__include: set to `True` to display a GUI element for this filter in the basic style interface.
        :param field__call_target: the factory to create a `Field` for the basic GUI, for example `Field.choice`. Default: `Field`
        """
        super(Filter, self).__init__(**kwargs)

    def on_refine_done(self):
        if 'choice' in getattr(self, 'iommi_shortcut_stack', []):
            assert (
                self.iommi_namespace.get('choices') is not None
            ), 'To use Filter.choice, you must pass the choices list'

        model_field = self.model_field
        if model_field and model_field.remote_field:
            self.model = model_field.remote_field.model
        super(Filter, self).on_refine_done()

    def on_bind(self) -> None:
        if self.attr is MISSING:
            self.attr = self._name

        # Not strict evaluate on purpose
        self.model = evaluate(self.model, **self.iommi_evaluate_parameters())

        if self.model and self.include and self.attr:
            try:
                self.search_fields = get_search_fields(model=self.model)
            except NoRegisteredSearchFieldException:
                self.search_fields = ['pk']

    def own_evaluate_parameters(self):
        return dict(filter=self)

    def __html__(self, *, render=None):
        assert False, (
            "Filters aren't rendered directly. You either render the Field corresponding to the filter, "
            "or the entire Query object."
        )

    @staticmethod
    @refinable
    def query_operator_to_q_operator(op: str) -> str:
        return Q_OPERATOR_BY_QUERY_OPERATOR[op]

    @staticmethod
    @refinable
    def parse(string_value, **_):
        return string_value

    @staticmethod
    @refinable
    def value_to_q(filter, op, value_string_or_f, **_) -> Q:
        if filter.attr is None:
            return Q()
        negated = False
        if op in ('!=', '!:'):
            negated = True
            op = op[1:]
        is_str = isinstance(value_string_or_f, str)
        if is_str and value_string_or_f.lower() == 'null':
            r = Q(**{filter.attr: None})
        else:
            if is_str:
                value_string_or_f = filter.parse(filter=filter, string_value=value_string_or_f, op=op)
            r = Q(**{filter.attr + '__' + filter.query_operator_to_q_operator(op): value_string_or_f})
        if negated:
            return ~r
        else:
            return r

    @classmethod
    def from_model(cls, model, model_field_name=None, model_field=None, **kwargs):
        return member_from_model(
            cls=cls,
            model=model,
            factory_lookup=_filter_factory_by_django_field_type,
            model_field_name=model_field_name,
            model_field=model_field,
            defaults_factory=lambda model_field: {},  # TODO: this is wrong! but base_defaults_factory doesn't work either.. there's no display_name on Filter
            **kwargs,
        )

    @classmethod
    @with_defaults(
        field__call_target__attribute='text',
        query_operator_for_field=':',
    )
    def text(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults
    def textarea(cls, **kwargs):
        return cls.text(**kwargs)

    @classmethod
    @with_defaults(
        query_operator_to_q_operator=case_sensitive_query_operator_to_q_operator,
    )
    def case_sensitive(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='choice',
    )
    def choice(cls, **kwargs):
        """
        Field that has one value out of a set.
        """
        instance = cls(**kwargs)
        instance = instance.refine(
            Prio.shortcut,
            field__choices=kwargs.get('choices'),
        )
        return instance

    @classmethod
    @with_defaults(
        field__call_target__attribute='multi_choice',
    )
    def multi_choice(cls, **kwargs):
        """
        Field that has one value out of a set.
        """
        instance = cls(**kwargs)
        instance = instance.refine(
            Prio.shortcut,
            field__choices=kwargs.get('choices'),
        )
        return instance

    @classmethod
    @with_defaults(
        field__call_target__attribute='choice_queryset',
        query_operator_to_q_operator=lambda op: 'exact',
        value_to_q=choice_queryset_value_to_q,
        is_valid_filter=choice_queryset__is_valid_filter,
    )
    def choice_queryset(cls, choices: QuerySet = None, **kwargs):
        """
        Field that has one value out of a set.
        """
        if 'model' not in kwargs:
            assert isinstance(
                choices, QuerySet
            ), 'The convenience feature to automatically get the parameter model set only works for QuerySet instances'
            kwargs['model'] = choices.model

        setdefaults_path(
            kwargs,
            dict(
                field__choices=choices,
                field__model=kwargs['model'],
                choices=choices,
            ),
        )
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='multi_choice_queryset',
    )
    def multi_choice_queryset(cls, **kwargs):
        return cls.choice_queryset(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='boolean',
        parse=bool_parse,
        unary=True,
        query_operator_to_q_operator=boolean__query_operator_to_q_operator,
    )
    def boolean(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='boolean_tristate',
        parse=boolean_tristate__parse,
        query_operator_to_q_operator=boolean__query_operator_to_q_operator,
        unary=True,
    )
    def boolean_tristate(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='number',
        query_operator_to_q_operator=case_sensitive_query_operator_to_q_operator,
    )
    def number(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='integer',
        parse=int_parse,
    )
    def integer(cls, **kwargs):
        return cls.number(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='float',
        parse=float_parse,
    )
    def float(cls, **kwargs):
        return cls.number(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='url',
    )
    def url(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='time',
        parse=time_parse,
    )
    def time(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='datetime',
        parse=datetime_parse,
        extra_evaluated__is_tz_aware=lambda **_: settings.USE_TZ,
    )
    def datetime(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='date',
        parse=date_parse,
    )
    def date(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='email',
    )
    def email(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='decimal',
    )
    def decimal(cls, **kwargs):
        return cls.number(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='file',
    )
    def file(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='duration',
    )
    def duration(cls, **kwargs):
        return cls.text(**kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='foreign_key',
    )
    def foreign_key(cls, model_field, **kwargs):
        setdefaults_path(
            kwargs,
            choices=model_field.foreign_related_fields[0].model.objects.all(),
        )
        return cls.choice_queryset(model_field=model_field, **kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='foreign_key_reverse',
    )
    def foreign_key_reverse(cls, *, model_field, **kwargs):
        setdefaults_path(
            kwargs,
            choices=model_field.remote_field.model.objects.all(),
            extra__django_related_field=True,
        )
        return cls.multi_choice_queryset(model_field=model_field, **kwargs)

    @classmethod
    @with_defaults(
        field__call_target__attribute='many_to_many',
    )
    def many_to_many(cls, model_field, **kwargs):
        setdefaults_path(
            kwargs,
            choices=model_field.remote_field.model.objects.all(),
            extra__django_related_field=True,
        )
        return cls.multi_choice_queryset(model_field=model_field, **kwargs)

    @classmethod
    @with_defaults
    def many_to_many_reverse(cls, model_field, **kwargs):
        return cls.many_to_many(model_field=model_field, **kwargs)


Filter.value_to_q.iommi_needs_attr = True


class StringValue(str):
    pass


def default_endpoint__errors(query, **_):
    try:
        query.get_q()
        errors = query.form.get_errors()
        # These dicts contains sets that we don't want in the JSON response, so convert to list
        if 'fields' in errors:
            errors['fields'] = {x: list(y) for x, y in items(errors['fields'])}
        if 'global' in errors:
            errors['global'] = list(errors['global'])
        return errors
    except QueryException as e:
        return {'global': [str(e)]}


class QueryAutoConfig(AutoConfig):
    rows = Refinable()


class Advanced(Fragment):
    toggle: Namespace = Refinable()

    @with_defaults(
        toggle__call_target=Action,
        toggle__attrs__href='#',
        toggle__attrs__class__iommi_query_toggle_simple_mode=True,
        toggle__attrs={'data-advanced-mode': 'simple'},
    )
    def __init__(self, **kwargs):
        super(Advanced, self).__init__(**kwargs)

    def on_refine_done(self):
        self.toggle = self.toggle(
            _name='toggle',
            display_name=gettext('Switch to advanced search'),
        ).refine_done(parent=self)
        super(Advanced, self).on_refine_done()

    def on_bind(self) -> None:
        super(Advanced, self).on_bind()
        self.toggle = self.toggle.bind(parent=self)


@declarative(Filter, '_filters_dict', add_init_kwargs=False)
@with_meta
class Query(Part):
    # language=rst
    """
    Declare a query language. Example:

    .. code-block:: python

        class AlbumQuery(Query):
            year = Filter.integer()
            name = Filter()

        query_set = Album.objects.filter(
            AlbumQuery().bind(request=request).get_q()
        )
    """

    auto: Namespace = Refinable()
    form: Namespace = Refinable()
    advanced: Namespace = Refinable()
    model: Type[Model] = SpecialEvaluatedRefinable()
    rows = Refinable()
    template: Union[str, Template] = EvaluatedRefinable()
    form_container: Fragment = EvaluatedRefinable()

    member_class = Refinable()
    form_class = Refinable()

    # Filters need to be at the end to not steal the short names
    filters: Namespace = RefinableMembers()

    class Meta:
        member_class = Filter
        form_class = Form
        filters = EMPTY
        auto = EMPTY

    @with_defaults(
        endpoints__errors__func=default_endpoint__errors,
        form__attrs={'data-iommi-errors': lambda query, **_: query.endpoints.errors.iommi_path},
        form_container__call_target=Fragment,
        form_container__tag='span',
        form_container__attrs__class__iommi_query_form_simple=True,
        advanced__call_target=Advanced,
    )
    def __init__(self, **kwargs):
        super(Query, self).__init__(**kwargs)

    def on_refine_done(self):
        assert isinstance(self.filters, dict)

        if self.auto:
            auto = QueryAutoConfig(**self.auto).refine_done(parent=self)
            model, rows, filters_from_auto = self._from_model(
                model=auto.model,
                rows=auto.rows,
                include=auto.include,
                exclude=auto.exclude,
            )

            assert self.model is None, (
                "You can't use the auto feature and explicitly pass model. "
                "Either pass auto__model, or we will set the model for you from auto__rows"
            )

            if self.model is None:
                self.model = model

            if self.rows is None:
                self.rows = rows
        else:
            filters_from_auto = None

        self.model, self.rows = model_and_rows(self.model, self.rows)

        self.query_advanced_value = None
        self.query_error = None

        refine_done_members(
            self,
            name='filters',
            members_from_namespace=self.filters,
            members_from_declared=self.get_declared('_filters_dict'),
            members_from_auto=filters_from_auto,
            cls=self.get_meta().member_class,
        )

        self._on_refine_done_form()

        self.advanced = self.advanced(_name='advanced').refine_done(parent=self)
        self.form_container = self.form_container(_name='form_container').refine_done(parent=self)

        super(Query, self).on_refine_done()

    @dispatch(
        render__call_target=render_template,
    )
    def __html__(self, *, render=None):
        if not self.iommi_bound_members().filters.iommi_bound_members():
            return ''

        setdefaults_path(
            render,
            context=self.iommi_evaluate_parameters(),
            template=self.template,
        )

        return render(request=self.get_request())

    def _on_refine_done_form(self):
        field_class = self.get_meta().form_class.get_meta().member_class

        declared_fields = Struct()

        freetext_search_config = self.iommi_namespace.form.get('fields', {}).get(FREETEXT_SEARCH_NAME, {})
        if freetext_search_config is not None:
            declared_fields[FREETEXT_SEARCH_NAME] = setdefaults_path(
                Namespace(),
                freetext_search_config,
                _name=FREETEXT_SEARCH_NAME,
                call_target__cls=field_class,
                display_name=gettext('Search'),
                required=False,
                include=lambda query, **_: any(filter.freetext for filter in values(query.filters)),
                help__include=False,
            )

        for name, filter in items(self.iommi_namespace.filters):
            _orig_include = getattr_path(filter, 'field__include', not filter.freetext)
            declared_fields[name] = setdefaults_path(
                Namespace(
                    include=(
                        lambda query, field, _orig_include=_orig_include, **_: (
                            field.iommi_name() in query.filters
                            and evaluate_strict(_orig_include, **query.iommi_evaluate_parameters())
                        )
                    )
                    if getattr(filter, 'include', None) is not False
                    else False,
                ),
                filter.field,
                _name=name,
                call_target__cls=field_class,
                model_field=filter.model_field,
                attr=name if filter.attr is MISSING else filter.attr,
                help__include=False,
            )

        # Remove fields from the form that correspond to non-included filters
        declared_filters = self.iommi_namespace.filters
        for name, field in items(declared_fields):
            if name == FREETEXT_SEARCH_NAME:
                continue
            # We need to check if it's in declared_filters first, otherwise we remove any injected fields
            if name in declared_filters and name not in self.iommi_namespace.filters:
                declared_fields[name] = field.refine_from_query(include=False)

        form_args = self.form

        # noinspection PyCallingNonCallable
        self.form: Form = (
            self.get_meta()
            .form_class(
                **setdefaults_path(
                    Namespace(),
                    form_args,
                    _name='form',
                    fields=declared_fields,
                    attrs__method='get',
                    actions__submit=dict(
                        attrs={'data-iommi-filter-button': ''},
                        display_name=pgettext(context='verb', message='Filter'),
                    ),
                )
            )
            .refine_done(parent=self)
        )

    def on_bind(self) -> None:
        # Prevent the nested form from thinking it's a part of a nested form set up
        if 'form' in self._evaluate_parameters:
            del self._evaluate_parameters['form']

        bind_members(self, name='filters')

        request = self.get_request()
        self.query_advanced_value = request_data(request).get(self.get_advanced_query_param(), '') if request else ''

        bind_members(self, name='endpoints')

        bind_member(self, name='form')
        bind_member(self, name='advanced')
        bind_member(self, name='form_container')

        self.filter_name_by_query_name = {x.query_name: name for name, x in items(self.filters)}

    @staticmethod
    @refinable
    def filter(query, rows, **_):
        if query.form:
            q = None
            try:
                q = query.get_q()
            except QueryException:
                pass
            if q:
                rows = rows.filter(q)

        return query.invoke_callback(query.postprocess, rows=rows)

    @staticmethod
    @refinable
    def postprocess(rows, **_):
        return rows

    def own_evaluate_parameters(self):
        return dict(query=self)

    def get_advanced_query_param(self):
        return '-' + path_join(self.iommi_path, 'query')

    def parse_query_string(self, query_string: str) -> Q:
        assert self._is_bound, NOT_BOUND_MESSAGE
        query_string = query_string.strip()
        if not query_string:
            return Q()
        parser = self._create_grammar()
        try:
            tokens = parser.parseString(query_string, parseAll=True)
        except ParseException:
            raise QueryException('Invalid syntax for query')
        return self._compile(tokens)

    def _compile(self, tokens) -> Q:
        items = []
        for token in tokens:
            if isinstance(token, ParseResults):
                items.append(self._compile(token))
            elif isinstance(token, Q):
                items.append(token)
            elif token in ('and', 'or'):
                items.append(token)
        return self._rpn_to_q(self._tokens_to_rpn(items))

    @staticmethod
    def _rpn_to_q(tokens):
        stack = []
        for each in tokens:
            if isinstance(each, Q):
                stack.append(each)
            else:
                op = each
                # infix right hand operator is on the top of the stack
                right, left = stack.pop(), stack.pop()
                stack.append(left & right if op == 'and' else left | right)
        assert len(stack) == 1
        return stack[0]

    @staticmethod
    def _tokens_to_rpn(tokens):
        # Convert a infix sequence of Q objects and 'and/or' operators using
        # dijkstra shunting yard algorithm into RPN
        if len(tokens) == 1:
            return tokens
        result_q, stack = [], []
        for token in tokens:
            assert token is not None
            if isinstance(token, Q):
                result_q.append(token)
            elif token in PRECEDENCE:
                p1 = PRECEDENCE[token]
                while stack:
                    t2, p2 = stack[-1]
                    if p1 <= p2:
                        stack.pop()
                        result_q.append(t2)
                    else:  # pragma: no cover
                        break  # pragma: no mutate
                stack.append((token, PRECEDENCE[token]))
        while stack:
            result_q.append(stack.pop()[0])
        return result_q

    def _create_grammar(self):
        """
        Pyparsing implementation of a where clause grammar based on http://pyparsing.wikispaces.com/file/view/simpleSQL.py

        The query language is a series of statements separated by AND or OR operators and parentheses can be used to group/provide
        precedence.

        A statement is a combination of three strings "<filter> <operator> <value>" or "<filter> <operator> <filter>".

        A value can be a string, integer or a real(floating) number or a (ISO YYYY-MM-DD) date.

        An operator must be one of "= != < > >= <= !:" and are translated into django __lte or equivalent suffixes.
        See self.as_q

        Example
        something < 10 AND other >= 2015-01-01 AND (foo < 1 OR bar > 1)

        """
        quoted_string_excluding_quotes = QuotedString('"', escChar='\\').setParseAction(
            lambda token: StringValue(token[0])
        )
        and_ = Keyword('and', caseless=True)
        or_ = Keyword('or', caseless=True)
        binary_op = oneOf('=> =< = < > >= <= : != !:', caseless=True).setResultsName('operator')

        # define query tokens
        identifier = Word(alphas, alphanums + '_$-.').setName('identifier')
        raw_value_chars = alphanums + '_$-+/$%*;?@[]\\^`{}|~.'
        raw_value = Word(raw_value_chars, raw_value_chars).setName('raw_value')
        value_string = quoted_string_excluding_quotes | raw_value

        # Define a where expression
        where_expression = Forward()
        binary_operator_statement = (identifier + binary_op + value_string).setParseAction(self._binary_op_to_q)
        unary_operator_statement = (identifier | (Char('!') + identifier)).setParseAction(self._unary_op_to_q)
        free_text_statement = quotedString.copy().setParseAction(self._freetext_to_q)
        operator_statement = binary_operator_statement | free_text_statement | unary_operator_statement
        where_condition = Group(operator_statement | ('(' + where_expression + ')'))
        where_expression << where_condition + ZeroOrMore((and_ | or_) + where_expression)

        # define the full grammar
        query_statement = Forward()
        query_statement << Group(where_expression).setResultsName("where")
        return query_statement

    def _unary_op_to_q(self, token):
        if len(token) == 1:
            (filter_name,) = token
            value = 'true'
        else:
            (op, filter_name) = token
            value = 'false'
            if op != '!':  # pragma: no cover. You can't actually get here because you'll get a syntax error earlier
                raise QueryException(f'Unknown unary filter operator "{op}", available operators: !')

        filter = self.filters.get(filter_name.lower())
        if filter:
            if not filter.unary:
                raise QueryException(
                    f'"{filter_name}" is not a unary filter, you must use it like "{filter_name}=something"'
                )
            result = filter.invoke_callback(filter.value_to_q, op='=', value_string_or_f=value)
            return result
        raise QueryException(
            f'Unknown unary filter "{filter_name}", available filters: {", ".join(list(keys(self.filters)))}'
        )

    def _binary_op_to_q(self, token):
        """
        Convert a parsed token of filter_name OPERATOR filter_name into a Q object
        """
        assert self._is_bound, NOT_BOUND_MESSAGE
        query_name, op, value_string_or_filter_name = token

        pk_lookup = False
        if query_name.endswith('.pk'):
            query_name = query_name[: -len('.pk')]
            pk_lookup = True

        filter_name = self.filter_name_by_query_name.get(query_name)
        if filter_name is None:
            raise QueryException(
                f'Unknown filter "{query_name}", available filters: {list(keys(self.filter_name_by_query_name))}'
            )

        filter = self._lowercase_filters().get(filter_name.lower())
        if filter is None:
            raise QueryException(
                f'Unknown filter "{query_name}", available filters: {list(keys(self.filter_name_by_query_name))}'
            )

        if pk_lookup:
            if op != '=':
                raise QueryException('Only = is supported for primary key lookup')

            try:
                pk = int(value_string_or_filter_name)
            except ValueError:
                pk = value_string_or_filter_name  # pk might be non-int

            if filter.pk_lookup_to_q:
                return filter.invoke_callback(filter.pk_lookup_to_q, pk=pk)
            else:
                return Q(**{f'{filter.attr}__pk': pk})

        if (
            isinstance(value_string_or_filter_name, str)
            and not isinstance(value_string_or_filter_name, StringValue)
            and value_string_or_filter_name.lower() in self.filters
        ):
            value_string_or_f = F(self.filters[value_string_or_filter_name.lower()].attr)
        else:
            value_string_or_f = value_string_or_filter_name
        try:
            result = filter.invoke_callback(filter.value_to_q, op=op, value_string_or_f=value_string_or_f)
        except ValidationError as e:
            raise QueryException(f'{e.message}')
        if result is None:
            raise QueryException(f'Unknown value "{value_string_or_f}" for filter "{query_name}"')
        return result

    def _freetext_to_q(self, token):
        if all(not v.freetext for v in values(self.filters)):
            raise QueryException('There are no freetext filters available')
        assert len(token) == 1
        token = token[0].strip('"')

        return reduce(
            operator.or_,
            [
                Q(**{filter.attr + '__' + filter.query_operator_to_q_operator(':'): token})
                for filter in values(self.filters)
                if filter.freetext
            ],
        )

    def get_query_string(self):
        """
        Based on the data in the request, return the equivalent query string that you can use with parse_query_string() to create a query set.
        """
        form = self.form
        request = self.get_request()

        if request is None:
            return ''

        if request_data(request).get(self.get_advanced_query_param(), '').strip():
            return request_data(request).get(self.get_advanced_query_param())
        elif form.is_valid():

            def expr(field, is_list, value):
                if is_list:
                    return '(' + ' OR '.join([expr(field, is_list=False, value=x) for x in field.value]) + ')'
                return build_query_expression(filter=self.filters[field._name], value=value)

            result = [
                expr(field, field.is_list, field.value)
                for field in values(form.fields)
                if field._name != FREETEXT_SEARCH_NAME
                and field._name in self.filters
                and field.value not in (None, '', [])
            ]

            if FREETEXT_SEARCH_NAME in form.fields:
                freetext = form.fields[FREETEXT_SEARCH_NAME].value
                if freetext:
                    result.append(
                        '(%s)'
                        % ' or '.join(
                            [
                                f'{filter.query_name}:{to_string_surrounded_by_quote(freetext)}'
                                for filter in values(self.filters)
                                if filter.freetext
                            ]
                        )
                    )
                    assert result[-1] != '()'
            return ' and '.join(result)
        else:
            return ''

    def get_q(self):
        """
        Create a query set based on the data in the request.
        """
        try:
            return self.parse_query_string(self.get_query_string())
        except QueryException as e:
            self.query_error = str(e)
            raise

    def _lowercase_filters(self):
        if not hasattr(self, '_lowercase_filters_cache'):
            self._lowercase_filters_cache = {k.lower(): v for k, v in items(self.filters)}
        return self._lowercase_filters_cache

    @classmethod
    @dispatch()
    def filters_from_model(cls, **kwargs):
        return create_members_from_model(
            member_class=cls.get_meta().member_class,
            **kwargs,
        )

    @classmethod
    @dispatch()
    def _from_model(cls, *, rows=None, model=None, include=None, exclude=None):
        assert rows is None or isinstance(rows, QuerySet), (
            'auto__rows needs to be a QuerySet for filter generation to work. '
            'If it needs to be a lambda, provide a model with auto__model for filter generation, '
            'and pass the lambda as rows.'
        )

        model, rows = model_and_rows(model, rows)
        assert model is not None or rows is not None, "auto__model or auto__rows must be specified"
        filters = cls.filters_from_model(model=model, include=include, exclude=exclude)
        return model, rows, filters



import functools
import re
import typing
from contextlib import contextmanager

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import (
    Model,
)
from django.db.models.base import ModelBase
from django.db.models.query_utils import DeferredAttribute
from django.http import Http404

from iommi.base import items
from iommi.struct import Struct

_camel_to_snake_regex = re.compile(r'(?<!^)(?=[A-Z])')


class Decoder:
    def __init__(self, *lookups, decode=None):
        self.lookups = lookups
        self._decode = decode

    def decode(self, *, lookup, string, model, **kwargs):
        if self._decode is None:
            return model.objects.get(**{lookup: string})
        else:
            return self._decode(lookup=lookup, string=string, model=model, **kwargs)


class PathDecoder:
    def __init__(self, *, decode=None, model=None, name):
        if decode is None:
            def decode(string, _model=model, **_):
                return _model.objects.get(pk=string)

        self.decode = decode
        self.model = model
        self.name = name


_path_component_to_decode_data: typing.Dict[
    str,
    typing.Tuple[typing.Optional[typing.Type[Model]], str, typing.Optional[str], typing.Union[Decoder, PathDecoder]],
] = {}


_default_decoder = Decoder('pk', 'name')


def camel_to_snake(s):
    return _camel_to_snake_regex.sub('_', s).lower()


def register_explicit_path_decoding(**kwargs):
    registered_keys = []
    for key, definition in items(kwargs):
        if isinstance(definition, ModelBase):
            decoder = PathDecoder(
                model=definition,
                name=camel_to_snake(definition.__name__),
            )
        elif isinstance(definition, property):
            assert False, f'Got a property for {key}. Maybe you did Foo.pk? In that case write just Foo, or Foo.id.'
        elif isinstance(definition, DeferredAttribute):
            field = definition.field
            model = field.model
            attr = field.name
            decoder = PathDecoder(
                decode=lambda string, _model=model, _attr=attr, **_: _model.objects.get(**{attr: string}),
                name=camel_to_snake(model.__name__),
            )
        elif callable(definition):
            decoder = PathDecoder(
                decode=definition,
                name=key,
            )
        else:
            assert isinstance(definition, PathDecoder)
            decoder = definition

        _path_component_to_decode_data[key] = (None, key, None, decoder)
        registered_keys.append(key)

    @contextmanager
    def _unregister():
        try:
            yield
        finally:
            for key in registered_keys:
                del _path_component_to_decode_data[key]

    return _unregister()


def register_path_decoding(**kwargs):
    return register_explicit_path_decoding(**kwargs)


def decode_path_components(request, **kwargs):
    decoded_kwargs = {}
    decoded_keys = set()

    if _path_component_to_decode_data:
        for k, v in items(kwargs):
            decode_data = _path_component_to_decode_data.get(k, None)
            if decode_data is None:
                continue

            model, key, lookup, decoder = decode_data

            try:
                if isinstance(decoder, Decoder):
                    # deprecated path
                    obj = decoder.decode(
                        lookup=lookup,
                        string=v,
                        request=request,
                        model=model,
                        snake_name=key,
                        decoded_kwargs=decoded_kwargs,
                        kwargs=kwargs,
                    )
                else:
                    obj = decoder.decode(
                        key=key,
                        string=v,
                        request=request,
                        decoded_kwargs=decoded_kwargs,
                        kwargs=kwargs,
                    )
                    key = decoder.name
            except ObjectDoesNotExist:
                raise Http404()

            decoded_kwargs[key] = obj
            decoded_keys.add(k)

    if not hasattr(request, 'iommi_view_params'):
        request.iommi_view_params = Struct()
    request.iommi_view_params.update(kwargs)
    request.iommi_view_params.update(decoded_kwargs)

    return {
        **{k: v for k, v in items(kwargs) if k not in decoded_keys},
        **decoded_kwargs,
    }


def decode_path(f):
    @functools.wraps(f)
    def decode_path_wrapper(request, **kwargs):
        decoded_kwargs = decode_path_components(request, **kwargs)
        return f(request=request, **decoded_kwargs)

    return decode_path_wrapper



__version__ = '6.8.0'

from functools import wraps

import django
from django.core.exceptions import ImproperlyConfigured

from iommi._db_compat import (
    register_factory,
    setup_db_compat,
)
from iommi.action import Action
from iommi.asset import Asset
from iommi.base import MISSING
from iommi.edit_table import (
    EditColumn,
    EditTable,
)
from iommi.form import (
    Field,
    Form,
    register_field_factory,
)
from iommi.fragment import (
    Fragment,
    Header,
    html,
)
from iommi.from_model import (
    register_search_fields,
)
from iommi.menu import (
    Menu,
    MenuItem,
)
from iommi.page import (
    Page,
)
from iommi.part import Part
from iommi.query import (
    Filter,
    Query,
    register_filter_factory,
)
from iommi.sort_after import LAST
from iommi.style import (
    Style,
    register_style,
)
from iommi.table import (
    Column,
    Table,
    register_cell_formatter,
    register_column_factory,
)

setup_db_compat()


def render_if_needed(request, response):
    if isinstance(response, Part):
        return render_part(request, response)
    else:
        return response


def render_part(request, part: Part):
    try:
        if not part._is_bound:
            part = part.bind(request=request)
        return part.render_to_response()
    except Exception as e:
        filename, lineno = part._instantiated_at_info
        if filename is None:
            raise

        from iommi.synthetic_traceback import SyntheticException

        fake = SyntheticException(
            tb=[dict(filename=filename, f_lineno=lineno, function='<iommi declaration>', f_globals={}, f_locals={})]
        )

        raise e from fake


# noinspection PyPep8Naming
class middleware:
    def __init__(self, get_response):
        from django.db import connections

        self.get_response = get_response
        self.atomic_db_aliases = {db.alias for db in connections.all() if db.settings_dict['ATOMIC_REQUESTS']}

    def process_view(self, request, view_func, view_args, view_kwargs):
        request.iommi_not_atomic_for = getattr(view_func, '_non_atomic_requests', set())

    def __call__(self, request):
        response = self.get_response(request)

        if isinstance(response, Part):
            from django.db import transaction

            render = render_part
            for alias in self.atomic_db_aliases:
                if alias not in request.iommi_not_atomic_for:
                    render = transaction.atomic(using=alias)(render)
            return render(request, response)

        return response


def iommi_render(view):
    @wraps(view)
    def inner(request, *args, **kwargs):
        result = view(request, *args, **kwargs)
        return render_if_needed(request, result)

    return inner


__all__ = [
    'Action',
    'Asset',
    'Column',
    'EditTable',
    'EditColumn',
    'Field',
    'Fragment',
    'Form',
    'Fragment',
    'Header',
    'Menu',
    'MenuItem',
    'middleware',
    'Page',
    'Part',
    'Query',
    'Table',
    'Filter',
    'LAST',
    'MISSING',
    'register_factory',
    'register_field_factory',
    'register_filter_factory',
    'register_column_factory',
    'register_cell_formatter',
    'register_style',
    'register_search_fields',
    'Style',
    'html',
    'iommi_render',
]

if django.VERSION[:2] < (3, 2):
    default_app_config = 'iommi.apps.IommiConfig'


try:
    from django.conf import settings
except ImproperlyConfigured:
    pass
else:
    if 'iommi' not in settings.INSTALLED_APPS:
        raise Exception("You must add 'iommi' to INSTALLED_APPS")



import functools
from typing import Type
from urllib.parse import urlencode

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import (
    auth,
    messages,
)
from django.contrib.auth.hashers import check_password
from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import ValidationError
from django.db.models import Model
from django.http import (
    Http404,
    HttpResponseRedirect,
)
from django.shortcuts import resolve_url
from django.urls import (
    path,
    reverse,
)
from django.utils.safestring import mark_safe
from django.utils.translation import gettext

from iommi import (
    LAST,
    Field,
    Form,
    Fragment,
    Menu,
    MenuItem,
    Page,
    Table,
    html,
    render_part,
)
from iommi._web_compat import format_html
from iommi.base import (
    build_as_view_wrapper,
    items,
    values,
)
from iommi.declarative.dispatch import dispatch
from iommi.declarative.namespace import (
    EMPTY,
    Namespace,
    flatten,
    setdefaults_path,
)
from iommi.declarative.with_meta import with_meta
from iommi.refinable import Refinable
from iommi.shortcut import with_defaults
from iommi.struct import Struct

app_verbose_name_by_label = {
    config.label: config.verbose_name.replace('_', ' ') for config in values(django_apps.app_configs)
}

joined_app_name_and_model = {
    f'{app_name}_{model_name}'
    for app_name, models in items(django_apps.all_models)
    for model_name, model in items(models)
}


@with_meta
class Messages(Fragment):
    class Meta:
        tag = 'div'

    def on_bind(self) -> None:
        super().on_bind()
        ms = messages.get_messages(self.get_request())
        if ms:
            self.children.update(
                {
                    f'message{i}': Fragment(
                        tag='div',
                        text=f'{m}',
                    ).bind(parent=self)
                    for i, m in enumerate(ms)
                }
            )


def collect_config(module):
    try:
        __import__(module.__name__ + '.iommi_admin')
        config_module = module.iommi_admin
    except ImportError:
        return None

    try:
        meta = config_module.Meta
    except AttributeError:
        return None

    return {k: v for k, v in meta.__dict__.items() if not k.startswith('_')}


def read_config(f):
    @functools.wraps(f)
    def read_config_wrapper(self, *args, **kwargs):
        from django.apps import apps

        configs = []
        for app_name, app in apps.app_configs.items():
            c = collect_config(app.module)
            if c is not None:
                configs.append(c)

        return f(self, *args, **Namespace(*configs, kwargs))

    return read_config_wrapper


@with_meta  # we need @with_meta again here to make sure this constructor gets all the meta arguments first
class Admin(Page):
    class Meta:
        iommi_style = 'bootstrap'
        table_class = Table
        form_class = Form
        apps = EMPTY
        parts = EMPTY

    model: Type[Model] = Refinable()
    instance: Model = Refinable()
    app_name: str = Refinable()
    model_name: str = Refinable()
    operation: str = Refinable()
    table_class: Type[Table] = Refinable()
    form_class: Type[Form] = Refinable()

    # Global configuration on apps level
    apps: Namespace = Refinable()

    menu = Menu(
        sub_menu=dict(
            root=MenuItem(
                url=lambda admin, **_: reverse('iommi.Admin.all_models', current_app=admin.app_name),
                display_name=gettext('iommi administration'),
            ),
            change_password=MenuItem(
                url=lambda admin, **_: reverse(Auth.change_password, current_app=admin.app_name),
                display_name=gettext('Change password'),
            ),
            logout=MenuItem(
                url=lambda admin, **_: reverse(Auth.logout, current_app=admin.app_name),
                display_name=gettext('Logout'),
            ),
        ),
    )

    @read_config
    @with_defaults(
        apps__auth_user__include=True,
        apps__auth_group__include=True,
        parts__messages=Messages(),
        parts__list_auth_user=dict(
            auto__include=['username', 'email', 'first_name', 'last_name', 'is_staff', 'is_active', 'is_superuser'],
            columns=dict(
                username__filter=dict(
                    include=True,
                    freetext=True,
                ),
                email__filter=dict(
                    include=True,
                    freetext=True,
                ),
                first_name__filter=dict(
                    include=True,
                    freetext=True,
                ),
                last_name__filter=dict(
                    include=True,
                    freetext=True,
                ),
                is_staff__filter__include=True,
                is_active__filter__include=True,
                is_superuser__filter__include=True,
            ),
        ),
    )
    def __init__(self, parts, apps, **kwargs):
        # Validate apps params
        for k in apps.keys():
            assert (
                k in joined_app_name_and_model
            ), f'{k} is not a valid app/model key.\n\n' f'Valid keys:\n    ' + '\n    '.join(
                sorted(joined_app_name_and_model)
            )
        super(Admin, self).__init__(parts=parts, apps=apps, **kwargs)

    def refine_with_params(self, app_name: str = None, model_name: str = None, pk: str = None):
        refined_admin = self.refine(app_name=app_name, model_name=model_name)

        model = django_apps.all_models[app_name][model_name] if app_name and model_name else None
        try:
            instance = model.objects.get(pk=pk) if pk is not None else None
        except model.DoesNotExist:
            raise Http404()

        if model is not None and instance is None:
            refined_admin = refined_admin.refine(model=model, parts__table__auto__model=model)

        if instance is not None:
            refined_admin = refined_admin.refine(instance=instance, parts__form__auto__instance=instance)

        return refined_admin

    def as_view(self):
        def admin_view(request, *args, **kwargs):
            app_name = kwargs.pop('app_name', None)

            if not getattr(request, 'user', None) or not request.user.is_authenticated:
                return HttpResponseRedirect(
                    f'{reverse(Auth.login, current_app=app_name)}?{urlencode(dict(next=request.path))}'
                )

            final_page = self.refine_with_params(
                app_name=app_name,
                model_name=kwargs.pop('model_name', None),
                pk=kwargs.pop('pk', None),
            ).refine_done()

            if not self.has_permission(
                request, instance=final_page.instance, model=final_page.model, operation=final_page.operation
            ):
                raise Http404()

            view = build_as_view_wrapper(final_page)

            return view(request, *args, **kwargs)

        return admin_view

    def on_refine_done(self):
        part_name = ''
        part_name += self.operation
        if self.app_name:
            part_name += '_' + self.app_name
        if self.model_name:
            part_name += '_' + self.model_name

        table = self.parts.get('table')
        if table is not None:
            setdefaults_path(self.parts, **{part_name: flatten(table)})
            self.parts.table = None

        form = self.parts.get('form')
        if form is not None:
            setdefaults_path(self.parts, **{part_name: flatten(form)})
            self.parts.form = None

        def should_throw_away(k, v):
            if k == part_name:
                return False

            if k == 'all_models':
                return True

            prefix_blacklist = [
                'list_',
                'delete_',
                'create_',
                'edit_',
            ]
            for prefix in prefix_blacklist:
                if k.startswith(prefix):
                    return True

            return False

        self.parts = {
            # Arguments that are not for us needs to be thrown on the ground
            k: None if should_throw_away(k, v) else v
            for k, v in items(self.parts)
        }

        super(Admin, self).on_refine_done()

    @staticmethod
    def has_permission(request, operation, model=None, instance=None):
        if request.user.is_superuser:
            return True

        if not request.user.is_staff:
            return False

        if model is None:
            return True

        action = {
            'create': 'add',
            'edit': 'change',
            'list': 'view',
        }.get(operation, operation)

        return request.user.has_perm(f'{model._meta.app_label}.{action}_{model._meta.model_name}')

    def own_evaluate_parameters(self):
        return dict(admin=self, **super(Admin, self).own_evaluate_parameters())

    @classmethod
    @with_defaults(
        operation='all_models',
    )
    def all_models(cls, table=None, **kwargs):
        def rows(admin, request, included_filter=False, **_):
            for app_name, models in items(django_apps.all_models):
                has_yielded_header = False

                for model_name, model in sorted(items(models), key=lambda x: x[1]._meta.verbose_name_plural):
                    key = f'{app_name}_{model_name}'
                    included = admin.apps.get(key, {}).get('include', False)
                    if included == included_filter:
                        continue

                    if not cls.has_permission(request, instance=None, model=model, operation='view'):
                        continue

                    if not has_yielded_header:
                        yield Struct(
                            name=app_verbose_name_by_label[app_name],
                            verbose_app_name=app_verbose_name_by_label[app_name],
                            app_name=None,
                            model_name=None,
                            url=None,
                            format=lambda row, table, **_: (
                                html.h2(row.name, _name='invalid_name').bind(parent=table).__html__()
                            ),
                            key=None,
                        )
                        has_yielded_header = True

                    yield Struct(
                        verbose_app_name=app_verbose_name_by_label[app_name],
                        app_name=app_name,
                        model_name=app_name,
                        name=model._meta.verbose_name_plural.capitalize(),
                        url='{}/{}/'.format(app_name, model_name),
                        format=lambda row, **_: row.name,
                        key=key,
                    )

        table = setdefaults_path(
            Namespace(),
            table if table is not None else {},
            title=gettext('All models'),
            call_target__cls=cls.get_meta().table_class,
            call_target__attribute='div',
            sortable=False,
            rows=rows,
            page_size=None,
            columns__name=dict(
                cell__url=lambda row, **_: row.url,
                display_name='',
                cell__format=lambda row, **format_kwargs: row.format(row=row, **format_kwargs),
            ),
        )

        add_models = setdefaults_path(
            Namespace(),
            include=settings.DEBUG,
            call_target__cls=cls.get_meta().table_class,
            sortable=False,
            rows=functools.partial(rows, included_filter=True),
            page_size=None,
            columns__conf=(
                cls.get_meta()
                .table_class.get_meta()
                .member_class(
                    cell__value=lambda row, **_: f'apps__{row.key}__include = True' if row.key else '',
                    cell__format=lambda value, **_: (
                        format_html(
                            '''
                                {}
                                <span style="margin-left: 1rem" onclick="navigator.clipboard.writeText('{}')">
                                    <i class="fa fa-copy"></i>
                                </span>
                            ''',
                            value,
                            value,
                        )
                        if value
                        else ''
                    ),
                    # cell__url=lambda row, **_: f'{row.url}add_model/' if row.key else None,
                    after=LAST,
                )
            ),
            columns__name=dict(
                display_name='',
                cell__format=lambda row, **format_kwargs: row.format(row=row, **format_kwargs),
            ),
        )

        return cls(
            parts=dict(
                all_models=table,
                add_models_title=html.h1(gettext('Add models to admin'), include=settings.DEBUG),
                add_models_help_text=html.p(
                    format_html(
                        '''
                            Copy the conf value to the `Meta` class of an `iommi_admin.py` file.

                            Read <a href="https://docs.iommi.rocks/en/latest/admin.html#customization">the docs for admin customization</a> for more information.
                        ''',
                    ),
                    include=settings.DEBUG,
                ),
                add_models=add_models,
            ),
            **kwargs,
        )

    @classmethod
    @dispatch(
        operation='list',
    )
    def list(cls, table=None, **kwargs):
        table = setdefaults_path(
            Namespace(),
            table if table is not None else {},
            call_target__cls=cls.get_meta().table_class,
            columns=dict(
                select__include=True,
                edit=dict(
                    call_target__attribute='edit',
                    after=0,
                    cell__url=lambda row, **_: '%s/edit/' % row.pk,
                ),
                delete=dict(
                    call_target__attribute='delete',
                    after=LAST,
                    cell__url=lambda row, **_: '%s/delete/' % row.pk,
                    include=lambda request, table, **_: (
                        cls.has_permission(request, instance=None, model=table.model, operation='delete')
                    ),
                ),
            ),
            actions=dict(
                create=dict(
                    display_name=lambda page, **_: (
                        gettext('Create %(model_name)s') % dict(model_name=page.model._meta.verbose_name)
                    ),
                    attrs__href='create/',
                ),
            ),
            query_from_indexes=True,
            bulk__actions__delete__include=lambda request, table, **_: (
                cls.has_permission(request, instance=None, model=table.model, operation='delete')
            ),
        )

        return cls(
            parts__header__children__link__attrs__href='../..',
            parts__table=table,
            **kwargs,
        )

    @classmethod
    @with_defaults
    def crud(cls, operation, form=None, **kwargs):
        def on_save(request, form, instance, **_):
            message = f'{form.model._meta.verbose_name.capitalize()} {instance} was ' + (
                'created' if form.extra.is_create else 'updated'
            )
            messages.add_message(request, messages.INFO, message, fail_silently=True)

        def on_delete(request, form, instance, **_):
            message = f'{form.model._meta.verbose_name.capitalize()} {instance} was deleted'
            messages.add_message(request, messages.INFO, message, fail_silently=True)

        form = setdefaults_path(
            Namespace(),
            form if form is not None else {},
            call_target__cls=cls.get_meta().form_class,
            call_target__attribute=operation,
            extra__on_save=on_save,
            extra__on_delete=on_delete,
            actions__submit__include=lambda request, form, **_: cls.has_permission(
                request, instance=None, model=form.model, operation=operation
            ),
            editable=lambda request, form, **_: (
                False
                if operation == 'delete'
                else cls.has_permission(request, instance=None, model=form.model, operation=operation)
            ),
        )

        return cls(
            parts__form=form,
            operation=operation,
            **kwargs,
        )

    @classmethod
    @with_defaults(
        parts__header__children__link__attrs__href='../../..',
    )
    def create(cls, **kwargs):
        return cls.crud(
            operation='create',
            **kwargs,
        )

    @classmethod
    @with_defaults(
        parts__header__children__link__attrs__href='../../../..',
    )
    def edit(cls, **kwargs):
        return cls.crud(
            operation='edit',
            **kwargs,
        )

    @classmethod
    @with_defaults(
        parts__header__children__link__attrs__href='../../../..',
    )
    def delete(cls, **kwargs):
        return cls.crud(
            operation='delete',
            **kwargs,
        )

    @classmethod
    def urls(cls):
        return Struct(
            urlpatterns=[
                path('', cls.all_models().as_view(), name='iommi.Admin.all_models'),
                path('<app_name>/<model_name>/', cls.list().as_view(), name='iommi.Admin.list'),
                path('<app_name>/<model_name>/create/', cls.create().as_view(), name='iommi.Admin.create'),
                path('<app_name>/<model_name>/<int:pk>/edit/', cls.edit().as_view(), name='iommi.Admin.edit'),
                path('<app_name>/<model_name>/<int:pk>/delete/', cls.delete().as_view(), name='iommi.Admin.delete'),
            ]
            + Auth.urls().urlpatterns
        )


class Auth:
    @classmethod
    def login(cls, request):
        return render_part(request, LoginPage())

    @classmethod
    def logout(cls, request):
        auth.logout(request)
        return HttpResponseRedirect(resolve_url(settings.LOGOUT_REDIRECT_URL or '/'))

    @classmethod
    def change_password(cls, request):
        return render_part(request, ChangePasswordPage())

    @classmethod
    def urls(cls):
        return Struct(
            urlpatterns=[
                path('login/', cls.login),
                path('logout/', cls.logout),
                path('change_password/', cls.change_password),
            ]
        )


class LoginForm(Form):
    username = Field(display_name=gettext('Username'))
    password = Field.password(display_name=gettext('Password'))

    class Meta:
        title = gettext('Login')

        @staticmethod
        def actions__submit__post_handler(form, **_):
            if form.is_valid():
                user = auth.authenticate(
                    username=form.fields.username.value,
                    password=form.fields.password.value,
                )

                if user is not None:
                    request = form.get_request()
                    auth.login(request, user)
                    return HttpResponseRedirect(request.GET.get('next', '/'))

                form.add_error(gettext('Unknown username or password'))


class LoginPage(Page):
    form = LoginForm()
    set_focus = html.script(
        mark_safe(
            'document.getElementById("id_username").focus();',
        )
    )


def current_password__is_valid(form, parsed_data, **_):
    return (
        (True, None)
        if check_password(parsed_data, form.get_request().user.password)
        else (False, gettext('Incorrect password'))
    )


def new_password__is_valid(form, parsed_data, **_):
    try:
        validate_password(parsed_data, form.get_request().user)
        return True, None
    except ValidationError as e:
        return False, ','.join(e)


def confirm_password__is_valid(form, parsed_data, **_):
    return (
        (True, None)
        if parsed_data == form.fields.new_password.value
        else (False, gettext('New passwords does not match'))
    )


class ChangePasswordForm(Form):
    class Meta:
        title = gettext('Change password')

        @staticmethod
        def actions__submit__post_handler(form, request, **_):
            if form.is_valid():
                user = request.user
                user.set_password(form.fields.new_password.value)
                user.save()
                return HttpResponseRedirect('..')

    current_password = Field.password(is_valid=current_password__is_valid, display_name=gettext('Current password'))
    new_password = Field.password(is_valid=new_password__is_valid, display_name=gettext('New password'))
    confirm_password = Field.password(is_valid=confirm_password__is_valid, display_name=gettext('Confirm password'))


class ChangePasswordPage(Page):
    form = ChangePasswordForm()
    set_focus = html.script(
        mark_safe(
            'document.getElementById("id_current_password").focus();',
        )
    )



from iommi.struct import (
    Frozen,
    Struct,
)


def _get_type_of_namespace(dict_value):
    if isinstance(dict_value, Namespace):
        return type(dict_value)
    else:
        return Namespace


class Namespace(Struct):
    """
    Namespace represents a structure of nested dicts. It behaves like a regular
    dictionary, with the added feature that values at nested levels can be set via
    `setitem_path` by providing a "path" of the form "<name1>__<name2>__<name3>",
    where the double underscores separate attribute names at increasing levels of
    depth.

    Attributes at nested levels can be retrieved either using "dotted" access
    such as `foo_namespace.a.b.c`, or via the helper function `getattr_path` by
    passing a double-underscored path like the example above.

    In addition, a Namespace can act like a function if it contains a toplevel
    item `{"call_target": f, ...}` where `f` is a callable. When the Namespace
    is called, `f` is called with the contents of the Namespace as keyword
    arguments (excluding f itself).
    """

    # noinspection PyMissingConstructor
    def __init__(self, *dicts, **kwargs):
        for mappings in dicts:
            for path, value in dict.items(mappings):
                self.setitem_path(path, value)
        for path, value in dict.items(kwargs):
            self.setitem_path(path, value)

    # Optimize attribute (and item) access by using native implementations.
    # Note: This comes with the price of `my_namespace['nonexisting']` raising AttributeError, not KeyError
    __getattribute__ = dict.__getitem__
    __missing__ = object.__getattribute__

    def setitem_path(self, path, value):
        key, delimiter, rest_path = path.partition('__')
        existing = Struct.get(self, key)

        if value is EMPTY:
            value = Namespace()

        if delimiter:
            if isinstance(existing, dict):
                type_of_namespace = _get_type_of_namespace(existing)
                self[key] = type_of_namespace(existing, {rest_path: value})
            elif callable(existing):
                self[key] = Namespace(dict(call_target=existing), {rest_path: value})
            else:
                # Unable to promote to Namespace, just overwrite
                self[key] = Namespace({rest_path: value})
        else:
            if existing is None:
                # This is a common case and checking for None is fast
                self[key] = value
            elif getattr(existing, 'shortcut', False):
                # Avoid merging Shortcuts
                self[key] = value
            elif isinstance(existing, dict):
                type_of_namespace = _get_type_of_namespace(existing)
                if isinstance(value, dict):
                    self[key] = type_of_namespace(existing, value)
                elif callable(value):
                    self[key] = type_of_namespace(existing, call_target=value)
                else:
                    # Unable to promote to Namespace, just overwrite
                    self[key] = value
            elif callable(existing):
                if isinstance(value, dict):
                    type_of_namespace = _get_type_of_namespace(value)
                    self[key] = type_of_namespace(dict(call_target=existing), value)
                else:
                    self[key] = value
            else:
                self[key] = value

    def __repr__(self):
        # Note: `repr` is called on any values in the namespace
        flattened_key_value_pairs = ", ".join(
            f'{k}={v!r}' for k, v in sorted(flatten_items(self), key=lambda x: x[0])
        )
        return f"{type(self).__name__}({flattened_key_value_pairs})"

    def __str__(self):
        # Note: `str` is called on any values in the namespace
        flattened_key_value_pairs = ", ".join(
            f'{k}={v}' for k, v in sorted(flatten_items(self), key=lambda x: x[0])
        )
        return "{}({})".format(type(self).__name__, flattened_key_value_pairs)

    def __call__(self, *args, **kwargs):
        params = Namespace(self, kwargs)

        try:
            call_target = params.pop('call_target')
        except KeyError as e:
            raise TypeError(
                'Namespace was used as a function, but no call_target was specified. The namespace is: %s' % self
            ) from e

        if isinstance(call_target, Namespace):
            if 'call_target' in call_target:
                # Override of the default
                call_target.pop('attribute', None)
                call_target.pop('cls', None)
            else:
                # The default
                attribute = call_target.get('attribute', None)
                if attribute is not None:
                    call_target = getattr(call_target.cls, attribute)
                else:
                    call_target = call_target.cls

        return call_target(*args, **params)


def func_from_namespace(namespace):
    call_target = namespace.get('call_target')
    func = None
    if isinstance(call_target, Namespace):
        if 'call_target' in call_target:
            return func_from_namespace(call_target)
        else:
            # The default
            attribute = call_target.get('attribute', None)
            cls = call_target.get('cls', None)
            if cls is not None:
                if attribute is not None:
                    func = getattr(call_target.cls, attribute)
                else:
                    func = call_target.cls
    else:
        func = call_target

    return func


class FrozenNamespace(Frozen, Namespace):
    pass


EMPTY = FrozenNamespace()


def flatten(namespace):
    return dict(flatten_items(namespace))


def flatten_items(namespace):
    def mappings(n, visited, prefix=''):
        for key, value in dict.items(n):
            path = prefix + key
            if isinstance(value, Namespace):
                if id(value) not in visited:
                    if value:
                        for mapping in mappings(value, visited=[id(value)] + visited, prefix=path + '__'):  # noqa: UP028
                            yield mapping
                    else:
                        yield path, Namespace()
            else:
                yield path, value

    return mappings(namespace, visited=[])


# The first argument has a funky name to avoid name clashes with stuff in kwargs
def setdefaults_path(__target__, *defaults, **kwargs):
    args = [kwargs] + list(reversed(defaults)) + [__target__]
    dict.update(__target__, Namespace(*args))
    return __target__


_MISSING = object()


def getattr_path(obj, path, default=_MISSING):
    """
    Get an attribute path, as defined by a string separated by '__'.
    getattr_path(foo, 'a__b__c') is roughly equivalent to foo.a.b.c but
    will short circuit to return None if something on the path is None.
    If no default value is provided AttributeError is raised if an attribute
    is missing somewhere along the path. If a default value is provided that
    value is returned.
    """
    if path == '':
        return obj
    current = obj
    parts = path.split('__')
    for name in parts:
        if default is _MISSING:
            try:
                current = getattr(current, name)
            except AttributeError as e:
                raise AttributeError(f"'{type(obj).__name__}' object has no attribute path '{path}', since {e}") from e

        else:
            current = getattr(current, name, _MISSING)
            if current is _MISSING:
                return default
        if current is None:
            return None
    return current


def setattr_path(obj, path, value):
    """
    Set an attribute path, as defined by a string separated by '__'.
    setattr_path(foo, 'a__b__c', value) is equivalent to "foo.a.b.c = value".
    """
    path = path.split('__')
    o = obj
    for name in path[:-1]:
        o = getattr(o, name)
    setattr(o, path[-1], value)
    return obj



import functools

from iommi.declarative import get_members
from iommi.declarative.namespace import Namespace
from iommi.refinable import Prio


# This is just a marker class for declaring shortcuts, and later for collecting them
class Shortcut(Namespace):
    shortcut = True


# decorator
def shortcut(f):
    f.shortcut = True
    return f


def is_shortcut(x):
    return getattr(x, 'shortcut', False)


def get_shortcuts_by_name(class_):
    return dict(get_members(class_, member_class=Shortcut, is_member=is_shortcut))


def with_defaults(__target__=None, **decorator_kwargs):
    def decorator(__target__):
        @functools.wraps(__target__)
        @shortcut
        def wrapper_for_with_defaults(*args, **kwargs):
            instance = __target__(*args, **kwargs)

            name = __target__.__name__
            if name == '__init__':
                args[0].refine(Prio.constructor, **decorator_kwargs)
            else:
                if decorator_kwargs:
                    instance = instance.refine(
                        Prio.shortcut,
                        **decorator_kwargs,
                    )

                shortcut_stack = [name] + getattr(instance, 'iommi_shortcut_stack', [])
                try:
                    instance.iommi_shortcut_stack = shortcut_stack
                except AttributeError:
                    pass

            return instance

        wrapper_for_with_defaults.__iommi_with_defaults_kwargs = decorator_kwargs
        return wrapper_for_with_defaults

    if __target__ is not None:
        return decorator(__target__)

    return decorator


def superinvoking_classmethod(f):
    @functools.wraps(f)
    def wrapper_for_superinvoking_classmethod(cls, *args, **kwargs):
        def super_classmethod_invoker(*args, **kwargs):
            parent_classmethod = None
            for parent_class in list(cls.mro())[1:]:
                candidate = vars(parent_class).get(f.__name__)
                if candidate is not None:
                    if parent_classmethod is None:
                        parent_classmethod = candidate
                    if candidate.__func__ == wrapper_for_superinvoking_classmethod:
                        # Infinite loop avoidance, we found ourselves.
                        parent_classmethod = None

            if parent_classmethod is None:
                raise TypeError(f'Unable to find parent class implementation of {cls.__name__}:{f.__name__}')

            undecorated_parent = parent_classmethod.__func__
            return undecorated_parent(cls, *args, **kwargs)

        return f(cls, *args, super_classmethod=super_classmethod_invoker, **kwargs)

    return wrapper_for_superinvoking_classmethod



from django.utils.translation import gettext

from iommi._web_compat import mark_safe
from iommi.fragment import html
from iommi.style import Style

bootstrap_icons = Style(
    root__assets__icons=html.link(
        attrs__rel="stylesheet",
        attrs__href="https://cdn.jsdelivr.net/npm/bootstrap-icons@1.7.1/font/bootstrap-icons.css",
    ),
    Column__shortcuts=dict(
        icon__extra=dict(
            icon_attrs__class={'bi': True},
            icon_prefix='bi-',
        ),
        edit__extra__icon='pencil-square',
        delete__extra__icon='trash',
        download__extra__icon='download',
        boolean__cell__format=lambda value, **_: mark_safe(
            f'<i class="bi bi-check-lg fs-3" title="{gettext("Yes")}"></i>'
        ) if value else '',
    ),
)



import inspect
import re
from glob import glob
from os import stat
from pathlib import Path
from textwrap import dedent
from typing import get_type_hints

from iommi import (
    MISSING,
    Part,
)
from iommi.base import items
from iommi.declarative import get_declared
from iommi.declarative.namespace import (
    Namespace,
    flatten,
)
from iommi.refinable import (
    EvaluatedRefinable,
    SpecialEvaluatedRefinable,
    is_refinable_function,
)
from iommi.shortcut import (
    get_shortcuts_by_name,
    is_shortcut,
)


def read_cookbook_links():
    result = []
    for filename in glob(str(Path(__file__).parent.parent / 'docs' / 'cookbook_*.rst')):
        with open(filename) as f:
            result.append((parse_cookbook_links(f.readlines()), Path(filename).stem))
    return result


def parse_cookbook_links(lines):
    link_marker = '.. _'
    anchors = [line.strip()[len(link_marker) :].rstrip(':') for line in lines if line.startswith(link_marker)]

    # TODO: validate that we only have anchors once

    return {x for x in anchors if not x.endswith('?')}


def validate_cookbook_links(cookbook_links):
    class_by_name = {x.__name__: x for x in get_default_classes()}

    for link, name in cookbook_links:
        if '.' in link:
            class_name, _, refinable_name = link.partition('.')
            assert class_name in class_by_name, f'{class_name} was not found in default_classes'
            getattr(class_by_name[class_name], refinable_name)


def get_default_classes():
    import iommi

    return [
        iommi.Table,
        iommi.Column,
        iommi.EditTable,
        iommi.EditColumn,
        iommi.Query,
        iommi.Filter,
        iommi.Form,
        iommi.fragment.Fragment,
        iommi.Field,
        iommi.Action,
        iommi.Page,
        iommi.Menu,
        iommi.MenuItem,
        iommi.Style,
        iommi.fragment.Header,
        iommi.Asset,
        iommi.middleware,
        # Private-ish APIs
        iommi.endpoint.Endpoint,
        iommi.part.Part,
        iommi.traversable.Traversable,
        iommi.member.Members,
        iommi.table.Cell,
        iommi.table.ColumnHeader,
        iommi.table.HeaderConfig,
        iommi.attrs.Attrs,
        iommi.table.ColumnHeader,
        iommi.fragment.Container,
    ]


def generate_api_docs_tests(directory, classes=None):  # pragma: no cover - this is tested by rtd anyway
    """
    Generate test files for declarative APIs

    :param directory: directory to write the .py files into
    :param classes: list of classes to generate tests for
    """
    if classes is None:
        classes = get_default_classes()

    doc_by_filename = _generate_tests_from_class_docs(classes=classes)  # pragma: no mutate
    for source_filename, filename, doc_generator in doc_by_filename:  # pragma: no mutate
        try:
            if stat(source_filename).st_mtime <= stat(directory / filename).st_mtime:
                continue
        except FileNotFoundError:
            pass

        doc = doc_generator()
        with open(directory / filename, 'w') as f2:  # pragma: no mutate
            f2.write(doc)  # pragma: no mutate


def get_docs_callable_description(c):
    if getattr(c, '__name__', None) == '<lambda>':
        import inspect

        return inspect.getsource(c).strip()
    return c.__module__ + '.' + c.__name__


def get_methods_by_type_by_name(class_):
    function_type = type(get_methods_by_type_by_name)
    ignore_list = ['get_declared', 'set_declared', 'get_meta']

    r = {
        k: v
        for k, v in class_.__dict__.items()
        if not k.startswith('_')
        and k not in ignore_list
        and not is_shortcut(getattr(class_, k))
        and not is_refinable_function(getattr(class_, k))
    }

    return {
        'Methods': {k: v for k, v in r.items() if isinstance(v, function_type)},
        'Static methods': {k: v for k, v in r.items() if isinstance(v, staticmethod)},
        'Class methods': {k: v for k, v in r.items() if isinstance(v, classmethod)},
    }


def docstring_param_dict(obj):
    doc = obj.__doc__
    if doc is None:
        return dict(text=None, params={})
    doc = dedent(doc)
    return dict(
        text=doc[: doc.find(':param')].strip() if ':param' in doc else doc.strip(),
        params=dict(re.findall(r":param (?P<name>\w+): (?P<text>.*)", doc)),
    )


def indent(levels, s):
    return (' ' * levels * 4) + s


def get_namespace(c):
    return Namespace(
        {
            k: getattr(c.__init__, '__iommi_with_defaults_kwargs', {}).get(k)
            for k, v in items(get_declared(c, 'refinable'))
        }
    )


def get_cookbook_name_by_refinable_name():
    cookbook_links = read_cookbook_links()
    validate_cookbook_links(cookbook_links)
    # TODO: validate that all cookbook links are actually linked somewhere. For example ".. _Column.cells_for_rows" isn't right now.
    cookbook_name_by_refinable_name = {}
    for links, name in cookbook_links:
        for link in links:
            cookbook_name_by_refinable_name[link] = name
    return cookbook_name_by_refinable_name


def _generate_tests_from_class_docs(classes):
    cookbook_name_by_refinable_name = get_cookbook_name_by_refinable_name()

    for c in classes:
        from io import StringIO

        f = StringIO()

        yield _generate_tests_from_class_doc(f, c, classes, cookbook_name_by_refinable_name)


def _generate_tests_from_class_doc(f, c, classes, cookbook_name_by_refinable_name):
    return (
        inspect.getfile(c),
        'test_doc__api_%s.py' % c.__name__,
        lambda: _generate_tests_from_class_doc_inner(f, c, classes, cookbook_name_by_refinable_name),
    )


def _generate_tests_from_class_doc_inner(f, c, classes, cookbook_name_by_refinable_name):
    def w(levels, s):
        f.write(indent(levels, s))
        f.write('\n')

    def section(level, title, indent=0):
        underline = {0: '=', 1: '-', 2: '^', 3: '+'}[level] * len(title)
        w(indent, title)
        w(indent, underline)
        w(indent, '')

    w(
        0,
        '''
# NOTE: this file is automatically generated

from iommi import *
from iommi.admin import Admin
from iommi.struct import Struct
from django.urls import (
    include,
    path,
)
import pytest
from django.db import models
from tests.helpers import req, user_req, staff_req, show_output
from docs.models import *

pytestmark = pytest.mark.django_db

@pytest.fixture(autouse=True)
def auto_use(big_discography):
    pass

request = req('get')


# language=rst
"""
    ''',
    )

    section(0, c.__name__, indent=0)

    class_doc = docstring_param_dict(c)
    constructor_doc = docstring_param_dict(c.__init__)

    if c.__base__ in classes:
        w(0, f'Base class: :doc:`{c.__base__.__name__}`')
    else:
        w(0, f'Base class: `{c.__base__.__name__}`')

    w(0, '')

    w(0, '"""')
    w(0, 'def test_base():')
    w(1, '# language=rst')
    w(1, '"""')

    _print_rst_or_python(class_doc['text'], w)

    w(1, '"""')
    w(0, '')
    w(1, '# language=rst')
    w(1, '"""')

    if constructor_doc['text']:
        if class_doc['text']:
            w(0, '')

        f.write(constructor_doc['text'])
        w(0, '')

    w(0, '')

    defaults = Namespace()
    for refinable, value in sorted(get_namespace(c).items()):
        if value not in (None, MISSING):
            defaults[refinable] = value

    def default_description(v):
        if callable(v) and not isinstance(v, Namespace):
            v = get_docs_callable_description(v)

            if 'lambda' in v:
                v = v[v.find('lambda') :]
                v = v.strip().strip(',').replace('\n', ' ').replace('  ', ' ')
        if isinstance(v, Part):
            v = v.bind()
        if v == '':
            v = '""'
        return v

    refinable_members = sorted(dict.items(get_namespace(c)))
    if refinable_members:
        section(1, 'Refinable members')
        type_hints = get_type_hints(c)
        for refinable, value in refinable_members:
            evaluated_marker = ''
            if isinstance(getattr(c, refinable, None), (EvaluatedRefinable, SpecialEvaluatedRefinable)):
                evaluated_marker = (
                    '  \N{NO-BREAK SPACE}\N{NO-BREAK SPACE}\N{NO-BREAK SPACE}  (:ref:`evaluated <evaluate>`)'
                )

            w(0, '')
            w(0, '* `' + refinable + '`' + evaluated_marker)

            docstring = getattr(getattr(c, refinable, None), '__doc__')
            if docstring:
                _print_rst_or_python(docstring, w, indent=1)
                w(0, '')
            elif constructor_doc['params'].get(refinable):
                w(1, constructor_doc['params'][refinable])
                w(0, '')
            type_hint = type_hints.get(refinable)
            if type_hint:
                name = str(type_hint)
                if name.startswith('typing.'):
                    name = name.replace('typing.', '')
                else:
                    name = type_hint.__name__

                if type_hint in classes:
                    w(1, f'Type: :doc:`{name}`')
                else:
                    w(1, f'Type: `{name}`')
                w(0, '')

            if refinable in defaults:
                w(1, f'Default: `{default_description(defaults.pop(refinable))}`')

            ref_name = f'{c.__name__}.{refinable}'
            if ref_name in cookbook_name_by_refinable_name:
                w(0, '')
                w(1, f'Cookbook: :ref:`{ref_name.lower()}`')
                w(0, '')

        w(0, '')

    assert not defaults

    shortcuts = get_shortcuts_by_name(c)
    if shortcuts:
        section(1, 'Shortcuts')

        for name, shortcut in sorted(shortcuts.items()):
            section(2, f'`{c.__name__}.{name}`')

            if shortcut.__doc__:
                foo = docstring_param_dict(shortcut)
                _print_rst_or_python(foo['text'], w)
                w(0, '')

                if foo['params']:
                    w(0, '')
                    section(3, 'Parameters')
                    for k, v in foo['params'].items():
                        w(0, f'* `{k}`')
                        w(1, f'* `{v}`')

                w(0, '')

            try:
                instance = shortcut()
            except Exception:
                pass  # todo For now...
            else:
                parents = instance.iommi_shortcut_stack[1:]
                if parents:
                    parent = parents[0]

                    w(0, f'Parent: {c.__name__}.{parent}_')
                    w(0, '')

            defaults = shortcut if isinstance(shortcut, dict) else getattr(shortcut, '__iommi_with_defaults_kwargs', {})
            if defaults:
                defaults = Namespace(defaults)
                section(3, 'Defaults')
                for k, v in items(flatten(defaults)):
                    v = default_description(v)
                    w(0, f'* `{k}`')
                    w(1, f'* `{v}`')
                w(0, '')

    for k, v in get_methods_by_type_by_name(c).items():
        methods = v
        if methods:
            section(1, k)

            for name, method in sorted(methods.items()):
                section(2, f'`{name}`')

                doc = getattr(c, name).__doc__
                if doc:
                    w(0, '')
                    _print_rst_or_python(doc, w)
                    w(0, '')

    w(1, '''"""''')
    return f.getvalue()


def _print_rst_or_python(doc, w, indent=0):
    if not doc:
        return
    in_code_block = False
    code_block_indent = None
    for line in dedent(doc).split('\n'):
        assert '\n' not in line
        if line.strip().startswith('.. code-block:: python'):
            w(1, '"""')

            in_code_block = True
        elif in_code_block:
            if code_block_indent is None:
                if line.strip():
                    code_block_indent = len(line) - len(line.lstrip(' '))
                w(0, line)
            else:
                if line.startswith(' ' * code_block_indent) or not line.strip():
                    w(0, line)
                else:
                    in_code_block = False
                    w(1, '# language=rst')
                    w(1, '"""')
                    w(1 + indent, line)
        else:
            w(0 + indent, line)
    if in_code_block:
        w(1, '# language=rst')
        w(1, '"""')
    w(0, '')



# Based on https://www.djangosnippets.org/snippets/186/

import cProfile
import os
import pstats
import subprocess
import sys
from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile

from django.http import StreamingHttpResponse
from django.utils.html import escape

from iommi._web_compat import (
    HttpResponse,
    settings,
)
from iommi.debug import src_debug_url_builder

MEDIA_PREFIXES = ['/static/']

_dot_search_paths = [
    '/usr/bin/dot',
    '/usr/local/bin/dot',
]


def get_dot_path():
    for p in _dot_search_paths:
        if os.path.exists(p):
            return p
    return None


def should_profile(request):
    disabled = getattr(request, 'profiler_disabled', True)
    is_staff = hasattr(request, 'user') and request.user.is_staff

    return ('_iommi_prof' in request.GET or '_iommi_prof' in request.POST) and ((not disabled and is_staff) or settings.DEBUG)


def strip_extra_path(s, token):
    if token not in s:
        return s
    pre, _, post = s.rpartition(' ')
    post = post[post.rindex(token) + len(token) :]
    return f'{pre} {post}'


class HTMLStats(pstats.Stats):
    def print_title(self):
        print('''
            <thead>
                <tr>
                    <th class="numeric"><a href="?_iommi_prof=ncalls">ncalls</a></th>
                    <th class="numeric"><a href="?_iommi_prof=tottime">tottime</a></th>
                    <th class="numeric">percall</th>
                    <th class="numeric"><a href="?_iommi_prof=cumtime">cumtime</a></th>
                    <th class="numeric">percall</th>
                    <th>function</th>
                    <th>filename</th>
                    <th>lineno</th>
                </tr>
            </thead>
        ''', file=self.stream)

    def print_stats(self, *amount):
        for filename in self.files:
            print(filename, file=self.stream)
        if self.files:
            print(file=self.stream)
        indent = ' ' * 8
        for func in self.top_level:
            print(indent, func[2], file=self.stream)

        print(indent, self.total_calls, "function calls", end=' ', file=self.stream)
        if self.total_calls != self.prim_calls:
            print("(%d primitive calls)" % self.prim_calls, end=' ', file=self.stream)
        print("in %.3f seconds" % self.total_tt, file=self.stream)
        print(file=self.stream)

        # this call prints...
        width, list = self.get_print_list(amount)

        print('<table>', file=self.stream)
        if list:
            self.print_title()
            limit = 280
            for func in list[:limit]:
                self.print_line(func)
            print(file=self.stream)
            print(file=self.stream)

        print('</table>', file=self.stream)
        return self

    def print_line(self, func):
        path, line_number, function_name = func

        base_dir = str(settings.BASE_DIR)
        should_bold = base_dir in path and '/site-packages/' not in path
        nice_path = path.replace(base_dir, '')
        nice_path = strip_extra_path(nice_path, '/site-packages')
        nice_path = strip_extra_path(nice_path, '/Python.framework/Versions')

        if should_bold:
            print('<tr class="own">', file=self.stream)
        else:
            print('<tr>', file=self.stream)

        def f8(x):
            return "%8.3f" % x

        cc, nc, tt, ct, callers = self.stats[func]
        c = str(nc)
        if nc != cc:
            c = c + '/' + str(cc)
        print(f'<td class="numeric">{c}</td>', file=self.stream)
        print(f'<td class="numeric">{f8(tt)}</td>', file=self.stream)
        if nc == 0:
            print('<td></td>', file=self.stream)
        else:
            print(f'<td>{f8(tt/nc)}</td>', file=self.stream)
        print(f'<td class="numeric">{f8(ct)}</td>', file=self.stream)
        if cc == 0:
            print('<td></td>', file=self.stream)
        else:
            print(f'<td class="numeric">{f8(ct/cc)}</td>', file=self.stream)

        if line_number and path:
            print(f'<td><a href="{src_debug_url_builder(path, line_number)}">{escape(function_name)}</a></td>', file=self.stream)
        else:
            print(f'<td>{escape(function_name)}</td>', file=self.stream)

        print(f'<td>{nice_path}</td>', file=self.stream)
        print(f'<td class="numeric">{line_number}</td>', file=self.stream)
        print('</tr>', file=self.stream)


class Middleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Disable profiling early on /media requests since touching request.user will add a
        # "Vary: Cookie" header to the response.
        request.profiler_disabled = False
        for prefix in MEDIA_PREFIXES:
            if request.path.startswith(prefix):
                request.profiler_disabled = True
                break

        if should_profile(request):
            prof = cProfile.Profile()
            prof.enable()
            request._iommi_prof = [prof]
        else:
            request._iommi_prof = []

        response = self.get_response(request)

        if request._iommi_prof:
            if isinstance(response, StreamingHttpResponse):
                # consume the entire streaming response, redirecting to stdout
                for line in response.streaming_content:
                    print(line.decode(), file=sys.__stdout__)

            response = HttpResponse()
            for prof in request._iommi_prof:
                prof.disable()


            s = StringIO()
            ps = HTMLStats(*request._iommi_prof, stream=s)

            prof_command = request.GET.get('_iommi_prof')

            if prof_command == 'graph':
                with NamedTemporaryFile() as stats_dump:
                    ps.stream = stats_dump
                    ps.dump_stats(stats_dump.name)

                    gprof2dot_path = Path(sys.executable).parent / 'gprof2dot'
                    if not gprof2dot_path.exists():
                        raise Exception('gprof2dot not found. Please install it to use the graph feature.')

                    with subprocess.Popen(
                        (sys.executable, gprof2dot_path, '-f', 'pstats', stats_dump.name), stdout=subprocess.PIPE
                    ) as gprof2dot:

                        response['Content-Type'] = 'image/svg+xml'

                        dot_path = get_dot_path()
                        if dot_path:
                            response.content = subprocess.check_output((dot_path, '-Tsvg'), stdin=gprof2dot.stdout)
                        else:
                            response['Content-Type'] = 'text/plain'
                            response['Content-Disposition'] = "attachment; filename=gprof2dot-graph.txt"
                            response.content = subprocess.check_output('tee', stdin=gprof2dot.stdout)

                        gprof2dot.wait()

            elif prof_command == 'snake':
                # noinspection PyPackageRequirements
                try:
                    import snakeviz  # noqa
                except ImportError:
                    return HttpResponse('You must `pip install snakeviz` to use this feature')

                with NamedTemporaryFile() as stats_dump:
                    ps.stream = stats_dump
                    ps.dump_stats(stats_dump.name)

                    subprocess.Popen(
                        [sys.executable, str(Path(sys.executable).parent / 'snakeviz'), stats_dump.name],
                        stdin=None,
                        stdout=None,
                        stderr=None,
                    )

                    # We need to wait a bit to give snakeviz time to read the file
                    from time import sleep

                    sleep(3)

                return HttpResponse(
                    'You should have gotten a new browser window with snakeviz opened to the profile data'
                )

            else:
                ps = ps.sort_stats(prof_command or 'cumulative')
                ps.print_stats()

                result = s.getvalue()

                # language=html
                start_html = '''
                    <style>
                        html {
                            font-family: monospace;
                            white-space: pre-line;
                        }

                        div, table {
                            white-space: normal;
                        }

                        td, th {
                            white-space: nowrap;
                            padding-right: 0.5rem;
                            color: #666;
                        }

                        th {
                            text-align: left;
                        }

                        .numeric {
                            text-align: right;
                        }

                        .own td {
                            font-weight: bold;
                            color: black;
                        }

                        @media (prefers-color-scheme: dark) {
                            html {
                                background-color: black;
                                color: #bbb;
                            }
                            td, th {
                                color: #888;
                            }

                            .own td {
                                color: white;
                            }

                            a {
                                color: #1d5aff;
                            }
                            a:visited {
                                color: #681dff;
                            }
                        }
                    </style>

                    <div>
                        <a href="?_iommi_prof=graph">graph</a>
                        <a href="?_iommi_prof=snake">snakeviz</a>
                    </div>

                    <p></p>
                '''

                response.content = start_html.strip() + result

                response['Content-Type'] = 'text/html'

        return response



import inspect
import json
from abc import abstractmethod
from typing import (
    Any,
    Dict,
    Union,
)

from django.core.serializers.json import DjangoJSONEncoder

from iommi._web_compat import (
    HttpResponse,
    HttpResponseBase,
    Template,
    get_template_from_string,
    render_template,
)
from iommi.base import (
    MISSING,
    NOT_BOUND_MESSAGE,
    items,
)
from iommi.debug import (
    get_instantiated_at_info,
    iommi_debug_on,
)
from iommi.declarative.dispatch import dispatch
from iommi.declarative.namespace import (
    EMPTY,
    Namespace,
)
from iommi.endpoint import (
    DISPATCH_PATH_SEPARATOR,
    Endpoint,
    InvalidEndpointPathException,
    perform_ajax_dispatch,
    perform_post_dispatch,
)
from iommi.member import (
    bind_members,
    refine_done_members,
)
from iommi.shortcut import with_defaults
from iommi.style import (
    get_style_object,
)
from iommi.traversable import (
    Traversable,
)

from ._web_compat import (
    QueryDict,
    settings,
    template_types,
)
from .refinable import (
    EvaluatedRefinable,
    Refinable,
    RefinableMembers,
    SpecialEvaluatedRefinable,
)
from .sort_after import sort_after


class Part(Traversable):
    """
    `Part` is the base class for parts of a page that can be rendered as html, and can respond to ajax and post.

    See the `howto <https://docs.iommi.rocks/en/latest/cookbook_parts_pages.html#parts-pages>`_ for example usages.
    """

    include: bool = SpecialEvaluatedRefinable()
    after: Union[int, str] = EvaluatedRefinable()
    extra: Dict[str, Any] = Refinable()
    # not EvaluatedRefinable because this is an evaluated container so is special
    extra_evaluated: Dict[str, Any] = Refinable()
    endpoints: Namespace = RefinableMembers()
    # Only the assets used by this part
    assets: Namespace = RefinableMembers()

    class Meta:
        extra = EMPTY

    @with_defaults(
        include=True,
    )
    def __init__(self, _collect_instantiated_at_info=True, **kwargs):
        super(Part, self).__init__(**kwargs)

        if _collect_instantiated_at_info:
            frame = inspect.currentframe()
            self._instantiated_at_info = get_instantiated_at_info(frame.f_back)

    def on_refine_done(self):
        from iommi.asset import Asset

        refine_done_members(self, name='endpoints', members_from_namespace=self.endpoints, cls=Endpoint)
        refine_done_members(self, name='assets', members_from_namespace=self.assets, cls=Asset)
        super().on_refine_done()

    @dispatch(
        render=EMPTY,
    )
    @abstractmethod
    def __html__(self, *, render=None):
        assert False, 'Not implemented'  # pragma: no cover, no mutate

    def __str__(self):
        assert self._is_bound, NOT_BOUND_MESSAGE
        return self.__html__()

    def bind(self, *, parent=None, request=None):
        result = super(Part, self).bind(parent=parent, request=request)
        if result is None:
            return None
        del self
        bind_members(result, name='endpoints')
        bind_members(result, name='assets', lazy=False)
        result.iommi_root()._iommi_collected_assets.update(result.assets)

        return result

    @dispatch
    def perform_dispatch(self, **kwargs):
        request = self.get_request()
        req_data = request_data(request)

        def dispatch_response_handler(r):
            if isinstance(r, HttpResponseBase):
                return r
            elif isinstance(r, Part):
                if not r._is_bound:
                    r = r.bind(request=request)
                return HttpResponse(render_root(part=r, **kwargs))
            else:
                return HttpResponse(json.dumps(r, cls=DjangoJSONEncoder), content_type='application/json')

        if request.method == 'GET':
            dispatch_prefix = DISPATCH_PATH_SEPARATOR
            dispatcher = perform_ajax_dispatch
            dispatch_error = 'Invalid endpoint path'

        elif request.method == 'POST':
            dispatch_prefix = '-'
            dispatcher = perform_post_dispatch
            dispatch_error = 'Invalid post path'

        else:  # pragma: no cover
            assert False  # This has already been checked in request_data()

        dispatch_commands = {key: value for key, value in items(req_data) if key.startswith(dispatch_prefix)}
        assert len(dispatch_commands) in (0, 1), 'You can only have one or no dispatch commands'
        if dispatch_commands:
            dispatch_target, value = next(iter(dispatch_commands.items()))
            try:
                result = dispatcher(root=self, path=dispatch_target, value=value)
            except InvalidEndpointPathException:
                if settings.DEBUG:
                    raise
                result = dict(error=dispatch_error)

            if result is not None:
                return dispatch_response_handler(result)
        else:
            if request.method == 'POST':
                assert False, 'This request was a POST, but there was no dispatch command present.'

        return None

    @dispatch
    def render_to_response(self, **kwargs):
        dispatch = self.perform_dispatch(**kwargs)
        if dispatch is not None:
            return dispatch

        response = HttpResponse(render_root(part=self, **kwargs))
        response.iommi_part = self
        return response

    def iommi_collected_assets(self):
        return sort_after(self.iommi_root()._iommi_collected_assets)


def get_title(part):
    from iommi import Header

    if isinstance(part, Header):
        for text in part.children.values():
            return text

    title = getattr(part, 'title', None)

    if title is None:
        parts = getattr(part, 'parts', None)
        if parts is not None:
            for p in parts.values():
                title = get_title(p)
                if title is not None:
                    break

    return title


@dispatch(
    render=EMPTY,
    context=EMPTY,
)
def render_root(*, part, context, **render):
    assert part._is_bound
    root_style = get_style_object(part)
    template_name = root_style.base_template
    content_block_name = root_style.content_block

    # Render early so that all the binds are forced before we look at all_assets,
    # since they are populated as a side-effect
    content = part.__html__(**render)

    assets = part.iommi_collected_assets()

    assert template_name, f"{root_style} doesn't have a base_template defined"
    assert content_block_name, f"{root_style} doesn't have a content_block defined"

    title = get_title(part)

    from iommi import Page
    from iommi.debug import iommi_debug_panel
    from iommi.fragment import Container

    request = part.get_request()

    context = dict(
        container=Container(_name='Container').refine_done(parent=part).bind(parent=part),
        content=content,
        title=title if title not in (None, MISSING) else '',
        iommi_debug_panel=(
            iommi_debug_panel(part)
            if iommi_debug_on() and '_iommi_disable_debug_panel' not in request.GET
            else ''
        ),
        iommi_language_code=getattr(request, 'LANGUAGE_CODE', settings.LANGUAGE_CODE),
        assets=assets,
        **(part.context if isinstance(part, Page) else {}),
        **context,
    )

    template_string = (
        '{% extends "'
        + template_name
        + '" %} {% block '
        + content_block_name
        + ' %}{{ iommi_debug_panel }}{{ content }}{% endblock %}'
    )
    return get_template_from_string(template_string).render(context=context, request=request)


PartType = Union[Part, str, Template]


def request_data(request):
    if request is None:
        return QueryDict()

    if request.method == 'POST':
        return request.POST
    elif request.method == 'GET':
        return request.GET
    else:
        assert False, f'unsupported request method {request.method}'


def as_html(*, request=None, part: PartType, context):
    if isinstance(part, str):
        return part
    elif isinstance(part, template_types):
        from django.template import RequestContext

        assert not isinstance(context, RequestContext)
        template = part
        return render_template(template=template, request=request, context=context)
    elif hasattr(part, '__html__'):
        return part.__html__()
    elif part is None:
        return ''
    else:
        return str(part)



from collections import defaultdict
from typing import (
    Optional,
    Type,
)

from django.db.models import QuerySet
from django.http import HttpResponseRedirect
from django.template import (
    Context,
    Template,
)
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy

from iommi.action import Action
from iommi.asset import Asset
from iommi.base import (
    MISSING,
    NOT_BOUND_MESSAGE,
    items,
    keys,
    values,
)
from iommi.declarative.dispatch import dispatch
from iommi.declarative.namespace import (
    EMPTY,
    Namespace,
    setdefaults_path,
)
from iommi.endpoint import (
    DISPATCH_PATH_SEPARATOR,
    path_join,
)
from iommi.form import (
    FULL_FORM_FROM_REQUEST,
    Field,
    Form,
)
from iommi.fragment import (
    Fragment,
)
from iommi.member import bind_member
from iommi.refinable import (
    Refinable,
    refinable,
)
from iommi.shortcut import with_defaults
from iommi.struct import Struct
from iommi.table import (
    Cell,
    Cells,
    Column,
    Table,
)


class EditCell(Cell):
    def get_path(self):
        return path_join(self.column.iommi_path, str(self.row.pk))

    def render_cell_contents(self):
        field = self.cells.get_field(self.column)

        if field:
            orig_attr = field.attr
            if self.cells.is_create_template:
                field.attr = None

            path = self.get_path()

            field.initial = MISSING
            field.raw_data = None
            field.value = None
            field.parsed_data = None
            field._errors = set()
            field.form.instance = self.row
            field._iommi_path_override = path
            field.input = field.iommi_namespace.input(_name='input')
            field.bind_from_instance()

            input_html = field.input.__html__()

            field.attr = orig_attr

            if self.table.edit_errors:
                errors = self.table.edit_errors.get(path)
                if errors:
                    return Template(
                        '{{ input_html }}<br><span class="text-danger"><ul class="errors">{% for error in errors %}<li>{{ error }}</li>{% endfor %}</ul></a>'
                    ).render(context=Context(dict(input_html=input_html, errors=errors)))

            return input_html
        else:
            return super().render_cell_contents()

    def on_refine_done(self):
        if self.cells.is_create_template:
            self.value = None
        super(EditCell, self).on_refine_done()


class EditCells(Cells):
    is_create_template = Refinable()

    class Meta:
        cell_class = EditCell
        is_create_template = False

    def get_field(self, column):
        table = self.iommi_evaluate_parameters()['table']
        if self.is_create_template:
            return table.create_form.fields.get(column.iommi_name(), None)
        else:
            return table.edit_form.fields.get(column.iommi_name(), None)

    def iter_editable_cells(self):
        table = self.iommi_evaluate_parameters()['table']
        for column in values(table.columns):
            field = self.get_field(column)
            if not field:
                continue
            yield self.cell_class(cells=self, column=column)


class EditColumn(Column):
    # language=rst
    """
    The column class for `EditTable`.
    """

    edit: Field = Refinable()

    def on_refine_done(self):
        super(EditColumn, self).on_refine_done()
        self.edit = None

    @classmethod
    @with_defaults(
        header__template='iommi/table/header.html',
        sortable=False,
        filter__is_valid_filter=lambda **_: (True, ''),
        filter__field__include=False,
        attr=None,
        display_name=gettext_lazy('Delete'),
        cell__attrs__class__delete=True,
        # language=js
        assets__fancy_delete=Asset(
            mark_safe(
                '''
                    $(document).ready(() => {
                        $('.edit_table_delete').click((event) => {
                            const checked = $(event.target).closest('tr').find('input')[0].checked;
                            $(event.target).closest('tr').find('input').prop("checked", !checked);
                            $(event.target).closest('tr')[0].style.opacity = checked ? "1.0" : "0.3";
                            event.preventDefault();
                            return false;
                        });
                    });
                '''
            ),
            tag='script',
        ),
    )
    def delete(cls, **kwargs):
        def cell__value(row, table, cells, column, **_):
            if isinstance(table.rows, QuerySet):
                row_id = row.pk
            else:
                # row_index is the visible row number
                # See selection() for the code that does the lookup
                row_id = cells.row_index
            button = Action.delete(display_name=column.display_name, attrs__class__edit_table_delete=True).bind()
            path = path_join(table.iommi_path, f'pk_delete_{row_id}')
            return mark_safe(f'{button.__html__()}<input style="display: none" type="checkbox" name="{path}" />')

        setdefaults_path(kwargs, dict(cell__value=cell__value))
        return cls(**kwargs)

    @classmethod
    @dispatch(
        edit=EMPTY,
    )
    def hardcoded(cls, **kwargs):
        assert (
            'parsed_data' in kwargs['edit']
        ), 'Specify a hardcoded value by specifying `edit__parsed_data` as a callable'
        return cls(**kwargs)


def edit_table__post_handler(table, request, **_):
    # 1. Validate all the fields
    table.edit_errors = defaultdict(set)
    table.create_errors = defaultdict(set)
    parsed_data = {}

    assert table.edit_form._is_bound
    assert table.create_form._is_bound
    table.edit_form.mode = FULL_FORM_FROM_REQUEST
    table.create_form.mode = FULL_FORM_FROM_REQUEST

    def validate(cells_iterator, form, errors):
        for cells in cells_iterator:
            if not isinstance(cells, EditCells):
                continue
            instance = cells.row
            form.instance = instance
            for cell in cells.iter_editable_cells():
                path = cell.get_path()
                field = form.fields[cell.column.iommi_name()]
                field._iommi_path_override = path
                field.input = field.iommi_namespace.input(_name='input')
                field.bind_from_instance()
                field_errors = field.get_errors()
                if field_errors:
                    errors[path] |= set(field_errors)
                else:
                    parsed_data[path] = field.value

    validate(table.cells_for_rows(), table.edit_form, table.edit_errors)
    validate(table.cells_for_rows_for_create(), table.create_form, table.create_errors)

    if table.edit_errors or table.create_errors:
        return None

    if isinstance(table.initial_rows, QuerySet):
        prefix = path_join(table.iommi_path, 'pk_delete_')
        table.bulk_queryset(prefix=prefix).delete()

    def save(cells_iterator, form):
        to_save = []
        for cells in cells_iterator:
            if not isinstance(cells, EditCells):
                continue
            instance = cells.row
            form.instance = instance
            attrs_to_save = []
            for cell in cells.iter_editable_cells():
                path = cell.get_path()
                value = parsed_data[path]
                field = form.fields[cell.column.iommi_name()]
                field._iommi_path_override = path
                if cells.is_create_template or (
                    field.invoke_callback(field.read_from_instance, instance=instance) != value
                ):
                    field.invoke_callback(field.write_to_instance, instance=instance, value=value)
                    if not field.extra.get('django_related_field', False):
                        attrs_to_save.append(field.attr)

            to_save.append((instance, attrs_to_save))

        to_save.sort(key=lambda x: abs(x[0].pk))
        for instance, attrs_to_save in to_save:
            if instance.pk is not None and instance.pk < 0:
                instance.pk = None
            if instance.pk is None:
                attrs_to_save = None
            instance.save(update_fields=attrs_to_save)

    save(table.cells_for_rows(), table.edit_form)
    save(table.cells_for_rows_for_create(), table.create_form)

    if 'post_save' in table.extra:
        table.invoke_callback(table.extra.post_save)

    return HttpResponseRedirect(request.META['HTTP_REFERER'])


class EditTable(Table):
    # language=rst
    """
    Describe an editable table. Example:

    .. code-block:: python

        table = EditTable(
            auto__model=Album,
            columns__name__edit__include=True,
        )

        # @test
        show_output(table)
        # @end
    """

    edit_errors = None
    create_errors = None
    edit_form: Form = Refinable()
    create_form: Form = Refinable()
    form_class: Type[Form] = Refinable()
    parent_form: Optional[Form] = Refinable()

    class Meta:
        form_class = Form
        member_class = EditColumn
        outer__tag = 'form'
        outer__attrs__enctype = 'multipart/form-data'
        outer__attrs__method = 'post'
        cells_class = EditCells
        actions__submit = dict(
            call_target__attribute='primary',
            display_name=gettext_lazy('Save'),
            post_handler=edit_table__post_handler,
            template=lambda table, **_: 'iommi/blank.html' if table.parent_form is not None else None,
        )
        actions__csrf = Action(
            tag='div',
            children__csrf=Fragment(
                template=lambda **_: Template('{% csrf_token %}'),
            ),
            attrs__style__display='none',
        )
        actions__add_row = Action.button(
            display_name=gettext_lazy('Add row'),
            attrs__onclick='iommi_add_row(this); return false',
        )
        actions_below = True
        edit_form = EMPTY
        create_form = EMPTY

        attrs = {
            'data-next-virtual-pk': '-1',
        }

        # language=js
        assets__edit_table_js = Asset.js(
            mark_safe(
                '''
                    function iommi_add_row(element) {
                        function find_for_siblings(s) {
                            while (s) {
                                let t = s.querySelector('table');
                                if (t) {
                                    return t;
                                }
                                s = s.previousElementSibling;
                            }
                            return null;
                        }

                        let table = null;
                        while (element.tagName !== 'FORM') {
                            element = element.parentNode;
                            let s = find_for_siblings(element);
                            if (s) {
                                table = s;
                                break;
                            }
                        }
                        if (!table) {
                            console.error('iommi: failed to find table!');
                            return;
                        }

                        let virtual_pk = parseInt(table.getAttribute('data-next-virtual-pk'), 10);
                        virtual_pk -= 1;
                        virtual_pk = virtual_pk.toString();
                        table.setAttribute('data-next-virtual-pk', virtual_pk);

                        let tmp = document.createElement('table');
                        tmp.innerHTML = table.getAttribute('data-add-template').replaceAll('#sentinel#', virtual_pk);
                        let y = tmp.querySelector('tr');
                        y.setAttribute('data-pk', virtual_pk)
                        table.querySelector('tbody').appendChild(y);
                        if (y.querySelector('.select2_enhance')) {
                            window.iommi.select2.initAll(y);
                        }
                    }
                '''
            )
        )

    def on_refine_done(self):
        super(EditTable, self).on_refine_done()

        fields = Struct()

        for name, column in items(self.iommi_namespace.columns):
            if not isinstance(column, EditColumn):
                continue
            if getattr(column, 'include', None) is False:
                continue

            edit_conf = column.iommi_namespace.get('edit', None)

            if not edit_conf:
                continue
            if getattr(column.edit, 'include', None) is False:
                continue

            if isinstance(edit_conf, dict):
                field = setdefaults_path(
                    Namespace(),
                    edit_conf,
                    model=self.model,
                    model_field_name=column.model_field_name,
                    attr=name if column.attr is MISSING else column.attr,
                )
            else:
                field = column.iommi_namespace.edit

            fields[name] = field

        auto = Namespace(self.auto)

        auto.pop('rows', None)
        if 'model' not in auto and 'rows' in auto:
            auto['model'] = auto.rows.model

        self.create_form = self.get_meta().form_class(
            **setdefaults_path(
                Namespace(),
                self.create_form,
                fields=fields,
                _name='create_form',
                auto=auto,
            )
        )

        if auto:
            auto.default_included = False

        self.edit_form = self.get_meta().form_class(
            **setdefaults_path(
                Namespace(),
                self.edit_form,
                fields=fields,
                _name='edit_form',
                auto=auto,
            )
        )

        declared_fields = self.edit_form.iommi_namespace.fields
        self.edit_form = self.edit_form.refine_defaults(fields=declared_fields).refine_done()
        self.create_form = self.create_form.refine_defaults(fields=declared_fields).refine_done()

    def on_bind(self) -> None:
        super(EditTable, self).on_bind()
        bind_member(self, name='edit_form')
        bind_member(self, name='create_form')

        # If this is a nested form register it with the parent, need
        # to do this early because is_target needs self.parent_form
        if self.iommi_parent() is not None:
            self.parent_form = self.iommi_parent().iommi_evaluate_parameters().get('form', None)
            if self.parent_form is not None:
                self.parent_form.nested_forms[self._name] = self
                self.outer.tag = None

        if self.model is not None:
            sentinel_row = self.model(pk='#sentinel#')
        else:
            sentinel_row = Struct(pk='#sentinel#', **{k: '' for k in keys(self.create_form.fields)})
        self.attrs['data-add-template'] = (
            self.cells_class(row=sentinel_row, row_index=-1, is_create_template=True, **self.row.as_dict())
            .bind(parent=self.create_form)
            .__html__()
        )

    def is_valid(self):
        return not self.edit_errors and not self.create_errors

    def cells_for_rows_for_create(self):
        """Yield a Cells instance for each create row sent from the client."""
        assert self._is_bound, NOT_BOUND_MESSAGE

        prefix = self.iommi_path + DISPATCH_PATH_SEPARATOR if self.iommi_path else ''
        delete_prefix = prefix + 'pk_delete_'

        def parse_virtual_pk(k):
            if not k.startswith(prefix) or k.startswith(delete_prefix):
                return None
            parts = k[len(prefix) :].split(DISPATCH_PATH_SEPARATOR)
            if len(parts) < 2:
                return None
            try:
                return int(parts[-1])
            except ValueError:
                return None

        pks = {parse_virtual_pk(k) for k in keys(self.get_request().POST)}

        virtual_pks = {k for k in pks if k is not None and k < 0}

        if not virtual_pks:
            return

        rows = [self.model(pk=pk) for pk in virtual_pks]

        for i, row in enumerate(rows):
            row = self.preprocess_row_for_create(table=self, row=row)
            assert row is not None, 'preprocess_row must return the row'
            yield self.cells_class(
                is_create_template=True,
                row=row,
                row_index=i,
                **self.row.as_dict(),
            ).bind(parent=self)

    def get_errors(self):
        return set()

    @staticmethod
    @refinable
    def preprocess_row_for_create(row, **_):
        return row



from django.utils.safestring import mark_safe

from iommi import Fragment
from iommi.asset import Asset
from iommi.style import (
    Style,
)
from iommi.style_base import (
    base,
    select2_enhanced_forms,
)
from iommi.style_font_awesome_4 import font_awesome_4

navbar_burger_click_js = Fragment(
    mark_safe(
        """\
<script>
    $(document).ready(function() {
          // Check for click events on the navbar burger icon
          $(".navbar-burger").click(function() {

              // Toggle the "is-active" class on both the "navbar-burger" and the "navbar-menu"
              $(".navbar-burger").toggleClass("is-active");
              $(".navbar-menu").toggleClass("is-active");

          });
    });
</script>
"""
    )
)

bulma_base = Style(
    base,
    root__assets=dict(
        css=Asset.css(
            attrs__href='https://cdn.jsdelivr.net/npm/bulma@0.9.1/css/bulma.min.css',
        ),
        navbar_burger_click_js=navbar_burger_click_js,
    ),
    sub_styles__horizontal=dict(
        Field=dict(
            attrs__class={
                'mr-4': True,
            },
            label__attrs__class={
                'mt-2': True,
                'mr-1': True,
            },
        ),
    ),
    Header__attrs__class={
        'title': True,
        'is-1': lambda fragment, **_: fragment.tag == 'h1',
        'is-2': lambda fragment, **_: fragment.tag == 'h2',
        'is-3': lambda fragment, **_: fragment.tag == 'h3',
        'is-4': lambda fragment, **_: fragment.tag == 'h4',
        'is-5': lambda fragment, **_: fragment.tag == 'h5',
        'is-6': lambda fragment, **_: fragment.tag == 'h6',
    },
    Container=dict(
        tag='div',
        attrs__class={
            'main': True,
            'container': True,
        },
    ),
    Field=dict(
        shortcuts=dict(
            boolean=dict(
                input__attrs__class__input=False,
                label__attrs__class__checkbox=True,
                label__attrs__class__label=False,
            ),
            textarea=dict(
                input__attrs__class__input=False,
                input__attrs__class__textarea=True,
            ),
            radio=dict(
                input__attrs__class__input=False,
            ),
        ),
        attrs__class__field=True,
        template='iommi/form/bulma/field.html',
        label__attrs__class__label=True,
        input__attrs__class__input=True,
        input__attrs__class={
            'is-danger': lambda field, **_: bool(field.errors),
        },
        errors__attrs__class={
            'is-danger': True,
            'help': True,
        },
        help__attrs__class=dict(
            help=True,
        ),
    ),
    Actions=dict(
        tag="div",
        attrs__class=dict(links=False, buttons=True),
    ),
    Action=dict(
        shortcuts=dict(
            # In bulma the most neutral button styling is button, which
            # gets you a button that's just an outline.
            button__attrs__class={
                'button': True,
            },
            delete__attrs__class={
                'is-danger': True,
            },
            primary__attrs__class={
                'is-primary': True,
            },
        ),
    ),
    Table={
        'attrs__class__table': True,
        'attrs__class__is-fullwidth': True,
        'attrs__class__is-hoverable': True,
    },
    Column=dict(
        shortcuts=dict(
            select=dict(
                header__attrs__title='Select all',
            ),
            number=dict(
                cell__attrs__class={
                    'has-text-right': True,
                },
                header__attrs__class={
                    'has-text-right': True,
                },
            ),
        ),
    ),
    Query__form_container=dict(
        tag='span',
        attrs__class={
            'is-horizontal': True,
            'field': True,
        },
    ),
    Menu=dict(
        attrs__class__navbar=True,
        tag='nav',
    ),
    MenuItem__a__attrs__class={'navbar-item': True},
    MenuItem__active_class='is-active',
    DebugMenu=dict(
        tag='aside',
        attrs__class={
            'navbar': False,
            'menu': True,
        },
    ),
    Paginator=dict(
        template='iommi/table/bulma/paginator.html',
        attrs__class__pagination=True,
    ),
    Errors__attrs__class={
        'help': True,
        'is-danger': True,
    },
)
bulma = Style(
    bulma_base,
    font_awesome_4,
    select2_enhanced_forms,
)



import itertools
import linecache
import logging
import re
import sys
import threading
from collections import defaultdict
from contextlib import contextmanager
from datetime import (
    date,
    datetime,
)
from logging import addLevelName
from time import monotonic

from django.conf import settings
from django.db import connections
from django.db.backends import utils as django_db_utils
from django.db.utils import DEFAULT_DB_ALIAS
from django.http import HttpResponse

from iommi._web_compat import format_html
from iommi.attrs import render_style
from iommi.thread_locals import (
    get_current_request,
    set_current_request,
)


def no_coloring(text, color=None, on_color=None, attrs=None):
    return text

try:
    from termcolor import colored
except ImportError:
    colored = no_coloring


log = logging.getLogger('db')
IOMMI_SQL_LOG = logging.INFO + 1
addLevelName(IOMMI_SQL_LOG, 'IOMMI_SQL_LOG')

SQL_DEBUG_LEVEL_ALL = 'all'
SQL_DEBUG_LEVEL_ALL_WITH_STACKS = 'stacks'
SQL_DEBUG_LEVEL_WORST = 'worst'
SQL_DEBUG_LEVEL_OFF = None

SQL_DEBUG_LEVELS = {
    SQL_DEBUG_LEVEL_ALL,
    SQL_DEBUG_LEVEL_ALL_WITH_STACKS,
    SQL_DEBUG_LEVEL_WORST,
    SQL_DEBUG_LEVEL_OFF,
}

assert getattr(settings, 'SQL_DEBUG', None) in SQL_DEBUG_LEVELS, f'SQL_DEBUG must be one of: {SQL_DEBUG_LEVELS}'


def linkify(s):
    from iommi.debug import src_debug_url_builder

    r = []
    for line in s.split('\n'):
        match = re.search(r'( *)File "(.*)", line (\d+), in (.*)', line)
        if match:
            starting_space = match.group(1)
            filename = match.group(2)
            lineno = int(match.group(3))
            function_name = match.group(4)
            r.append(
                format_html(
                    '{}File "<a href="{}">{}</a> ", line {}, in {}\n',
                    starting_space,
                    src_debug_url_builder(filename, lineno),
                    filename,
                    lineno,
                    function_name,
                )
            )

    return format_html('{}' * len(r), *r)


class Middleware:
    def __init__(self, get_response):
        self.get_response = get_response
        self.prof = None

    def __call__(self, request):
        set_current_request(request)
        request.iommi_start_time = datetime.now()
        sql_trace = request.GET.get('_iommi_sql_trace')

        if sql_trace is None:
            sql_trace = get_sql_debug()

        if sql_trace == '':
            sql_trace = SQL_DEBUG_LEVEL_WORST

        assert sql_trace in SQL_DEBUG_LEVELS

        old_state = get_sql_debug()
        set_sql_debug(sql_trace)
        try:
            request.iommi_sql_debug_log = []

            response = self.get_response(request)

            sql_debug_last_call(response)

            if '_iommi_sql_trace' not in request.GET:
                return response

            if not settings.DEBUG and not request.user.is_staff:
                return response

            if sql_trace is not None:
                iommi_sql_debug_log = getattr(request, 'iommi_sql_debug_log', None)
                if iommi_sql_debug_log is not None:
                    total_duration = float(sum(x['duration'] for x in iommi_sql_debug_log))
                    result = [
                        '''
                        <style>
                            a, span {
                                box-sizing: border-box;
                            }
                            @media (prefers-color-scheme: dark) {
                                html {
                                    background-color: black;
                                    color: #bbb;
                                }
                                b {
                                    color: white;
                                }
                            }
                            </style>
                        ''',
                        f'{len(iommi_sql_debug_log)} queries, {total_duration:.3} seconds total<br><br>',
                    ]

                    if total_duration:
                        color = '#4f4fff'
                        for i, x in enumerate(iommi_sql_debug_log):
                            proportion = x["duration"] / total_duration * 100
                            result.append(
                                f'<a href="#query_{i}" style="display: inline-block; height: 30px; width: {proportion}%; background-color: {color}; border-left: 1px solid black" title="{x["duration"]:.3}s"></a>'
                            )

                        result.append('<br><br>By group:<br>')

                        for k, group in itertools.groupby(iommi_sql_debug_log, key=lambda x: x['sql']):
                            group = list(group)
                            duration = sum(x["duration"] for x in group)
                            proportion = duration / total_duration * 100
                            k = k.replace('>', '&gt;')
                            result.append(
                                format_html(
                                    '<span style="display: inline-block; height: 30px; width: {}%; background-color: {}; border-left: 1px solid black" title="{} queries. Ran {}s, like {}"></span>',
                                    proportion,
                                    color,
                                    len(group),
                                    f'{duration:.3}',
                                    k,
                                )
                            )

                    result.append('<p></p><pre>')

                    for i, x in enumerate(iommi_sql_debug_log):
                        sql = x['sql']
                        if x['params']:
                            sql %= safe_unicode_literal(x['params'])

                        result.append(f'<a name="query_{i}">')
                        result.append(format_sql(sql, duration=x['duration']))
                        if sql_trace == SQL_DEBUG_LEVEL_ALL_WITH_STACKS:
                            result.append('\n\n')
                            result.append(linkify(x['stack']))
                        result.append('\n\n')
                    result.append('</pre>')
                    return HttpResponse(''.join(result))

            return response
        finally:
            set_sql_debug(old_state)


def colorize(text, fg, bold=False):
    styles = {}
    if fg:
        styles['color'] = fg
    if bold:
        styles['font-weight'] = 'bold'
    if styles:
        return format_html('<span style="{}">{}</span> ', render_style(styles), text)
    else:
        return text


def format_sql(text, width=60, duration=None, short_limit=120):
    base_colors = {
        'SELECT': 'green',
        'INSERT': 'magenta',
        'UPDATE': 'orange',
        'DELETE': 'red',
        'COMMIT': 'cyan',
        'BEGIN': 'cyan',
        'ROLLBACK': 'red',
    }
    # pretend all upper case tokens are keywords
    sql_keyword_re = re.compile(r'^[A-Z]+$')
    sql_token_sep_re = re.compile(r'\s+')

    def tokenize(text):
        tokens = re.split(sql_token_sep_re, text.strip().replace('`', ''))
        short = len(text) < short_limit

        _first = True
        column = 0
        fg = None
        for token in tokens:
            bold = False
            if sql_keyword_re.match(token):
                bold = True
                if _first and token in base_colors:
                    _first = False
                    fg = base_colors[token]
                if not short:
                    if token in ('AND', 'OR', 'LEFT', 'INNER', 'FROM', 'WHERE', 'ORDER', 'LIMIT'):
                        yield format_html('<span><br>&nbsp;</span>')
                        column = 2
                        if token in (
                            'AND',
                            'OR',
                        ):
                            yield format_html('&nbsp;')
                            column += 2
            yield colorize(token, fg=fg, bold=bold)

            column += len(token) + 1
            if column > width and token[-1] == ',':
                yield format_html('<span><br>{}</span>', format_html('&nbsp;' * 8))
                column = 8
        if duration:
            yield format_html(f"<span>-- [{duration:.3f}s]</span>")

    tokens = list(tokenize(text))
    return format_html('<span>' + ('{}' * len(tokens)) + '</span>', *tokens)


def safe_unicode_literal(obj):
    if obj is None:
        return 'NULL'

    if isinstance(obj, (list, tuple)):
        return tuple([safe_unicode_literal(x) for x in obj])
    elif isinstance(obj, (float, int)):
        return repr(obj)
    elif isinstance(obj, (date, datetime)):
        return repr(obj.isoformat())
    elif isinstance(obj, dict):
        return dict((k, safe_unicode_literal(v)) for k, v in obj.items())
    elif isinstance(obj, bytes):
        return repr(obj.decode(errors="replace"))
    else:
        return repr(obj)


state = threading.local()


def set_sql_debug(new_state, validate=True):
    if new_state == 'None':
        new_state = None
    if validate:
        assert new_state is None or new_state in SQL_DEBUG_LEVELS
    setattr(state, 'sql_debug', new_state)


def get_sql_debug():
    sentinel = object()
    result = getattr(state, 'sql_debug', sentinel)
    if result is not sentinel:
        return result
    return getattr(settings, 'SQL_DEBUG', SQL_DEBUG_LEVEL_WORST)


@contextmanager
def no_sql_debug():
    """
    Context manager to temporarily suspend sql logging.

    This is useful inside the sql debug implementation to avoid infinite recursion.
    """
    old_state = get_sql_debug()
    set_sql_debug(SQL_DEBUG_LEVEL_OFF)
    yield
    set_sql_debug(old_state)


def sql_debug(msg, **extra):
    if get_sql_debug():
        if 'sql' not in extra:
            # If we don't do this we'll end up with infinite recursion back here
            extra['sql'] = ''
        log.log(level=IOMMI_SQL_LOG, msg=msg, extra=extra)


def sql_debug_log_to_request(**data):
    request = get_current_request()
    if request is not None:
        try:
            request.iommi_sql_debug_log.append(data)
        except AttributeError:
            request.iommi_sql_debug_log = [data]

    level = get_sql_debug()
    if level is False or level == SQL_DEBUG_LEVEL_WORST:
        return
    sql_debug_trace_sql(**data)


def sql_debug_trace_sql(sql, params=None, **kwargs):
    if params:
        sql %= safe_unicode_literal(params)
    if len(sql) > 10000:
        sql = '%s... [%d bytes sql]' % (re.sub(r'[\x00-\x08\x0b-\x1f\x80-\xff].*', '.', sql[:10000]), len(sql))
    else:
        sql = re.sub(r'[\x00-\x08\x0b-\x1f\x80-\xff]', '.', sql)

    if get_sql_debug() == SQL_DEBUG_LEVEL_ALL_WITH_STACKS:
        sql_debug(kwargs['stack'], extra={})

    kwargs['sql'] = True
    sql_debug(sql, extra=kwargs)

    if get_sql_debug() == SQL_DEBUG_LEVEL_ALL_WITH_STACKS:
        sql_debug('\n', extra=kwargs)
    return sql


def sql_debug_total_time():
    try:
        log = get_current_request().iommi_sql_debug_log
    except AttributeError:
        return 0

    return sum(x['duration'] for x in log)


def get_line_cached_on_request(file_name, line):
    request = get_current_request()

    if not request:
        return linecache.getline(file_name, line)

    if not hasattr(request, '_iommi_line_cache'):
        request._iommi_line_cache = {}

    if file_name not in request._iommi_line_cache:
        request._iommi_line_cache[file_name] = linecache.getlines(file_name)

    if not request._iommi_line_cache[file_name]:
        return ''
    return request._iommi_line_cache[file_name][line - 1]


def format_clickable_filename(file_name, line, fn, extra=None):
    if not extra:
        if line is not None:
            extra = get_line_cached_on_request(file_name, line)

    if extra is None:
        extra = '<unknown>'
    if file_name is None:
        file_name = '<unknown>'
    if line is None:
        line = '<unknown>'
    if fn is None:
        fn = '<unknown>'

    return f'  File "{file_name}", line {line}, in {fn} => {extra.strip()}'.rstrip()


def sql_debug_format_stack_trace(frame):
    if frame is None:
        return None

    lines = []
    skip_template_code = False

    def skip_line(frame):
        filename = frame.f_code.co_filename
        return (
            '/lib/python' in filename
            or 'django/core' in filename
            or 'pydev/pydevd' in filename
            or 'gunicorn' in filename
            or 'iommi/iommi/declarative/dispatch.py' in filename
            or 'iommi/iommi/member.py' in filename
        )

    while frame:
        if skip_line(frame):
            frame = frame.f_back
            continue

        file_name, line, fn = frame.f_code.co_filename, frame.f_lineno, frame.f_code.co_name
        extra = ''
        if fn == '_resolve_lookup':
            f = frame
            while f.f_back and 'bit' not in f.f_locals:
                f = f.f_back
            if 'bit' in f.f_locals:
                extra = colored(f'(looking up: {f.f_locals["bit"]}) ', color='red')
        elif "django/template" in file_name:
            if skip_template_code:
                frame = frame.f_back
                continue
            skip_template_code = True
        elif skip_template_code:
            skip_template_code = False

        lines.append(format_clickable_filename(file_name, line, fn, extra))

        frame = frame.f_back

    stack = "\n".join(lines).rstrip()
    return stack


def sql_debug_last_call(response):
    request = get_current_request()
    if get_sql_debug() == SQL_DEBUG_LEVEL_WORST and hasattr(
        request, 'iommi_sql_debug_log'
    ):  # hasattr check because process_request might not be called in case of an early redirect
        stacks = defaultdict(list)
        for x in request.iommi_sql_debug_log:
            stacks[x['sql']].append(x)

        highscore = sorted([(len(logs), stack_trace, logs) for stack_trace, logs in stacks.items()])
        # Print the worst offenders
        number_of_offenders = getattr(settings, 'SQL_DEBUG_WORST_NUMBER_OF_OFFENDERS', 3)
        query_cutoff = getattr(settings, 'SQL_DEBUG_WORST_QUERY_CUTOFF', 4)
        num_suspicious = getattr(settings, 'SQL_DEBUG_WORST_SUSPICIOUS_CUTOFF', 3)
        for count, _, logs in highscore[-number_of_offenders:]:
            if count > num_suspicious:  # 3 times is ok-ish, more is suspicious
                sql_debug(f'------ {count} times: -------', bold=True)
                if hasattr(response, 'iommi_part'):
                    from iommi.debug import filename_and_line_num_from_part

                    filename, lineno = filename_and_line_num_from_part(response.iommi_part)
                    sql_debug('From source:')
                    sql_debug(
                        format_clickable_filename(filename, lineno, str(response.iommi_part._name)), sql_trace=True
                    )
                    sql_debug('With Stack:')
                sql_debug(logs[0]['stack'], sql_trace=True)
                for x in logs[:query_cutoff]:
                    sql_debug_trace_sql(**x)
                if len(logs) > query_cutoff:
                    sql_debug(f'... and {len(logs) - query_cutoff:d} more unique statements\n')
        queries_per_using = defaultdict(int)
        for x in request.iommi_sql_debug_log:
            queries_per_using[x['using']] += 1

        sql_debug(f'Total number of SQL statements: {sum(queries_per_using.values())}\n')

    if settings.DEBUG:
        total_sql_time = f" (sql time: {sql_debug_total_time():.3f}s)"
        duration = f' ({(datetime.now() - request.iommi_start_time).total_seconds():.3f}s)'
        sql_debug(
            msg=f'{request.META["REQUEST_METHOD"]} {request.get_full_path()} -> {response.status_code} {duration}{total_sql_time}'
        )
        sql_debug(f'{request.get_full_path()} -> {response.status_code}{duration}', fg='magenta')

    set_sql_debug(None)


class CursorDebugWrapper(django_db_utils.CursorWrapper):
    def _execute_and_log(self, *, f, **kwargs):
        if get_sql_debug() is not None:
            frame = sys._getframe().f_back.f_back
            while "django/db" in frame.f_code.co_filename:
                frame = frame.f_back
        else:
            frame = None

        start = monotonic()
        try:
            return f(**kwargs)
        finally:
            stop = monotonic()
            duration = stop - start
            sql_debug_log_to_request(
                stack=sql_debug_format_stack_trace(frame),
                duration=duration,
                rowcount=self.cursor.rowcount,
                using=self.db.alias,
                **kwargs,
            )

    def execute(self, sql, params=None):
        return self._execute_and_log(f=super().execute, sql=sql, params=params)

    def executemany(self, sql, param_list):
        return self._execute_and_log(f=super().executemany, sql=sql, param_list=param_list)

    def __iter__(self):
        return iter(self.cursor)

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            return getattr(self.cursor, attr)


def make_debug_cursor(self, cursor):
    return CursorDebugWrapper(cursor, self)


connections[DEFAULT_DB_ALIAS].__class__.make_debug_cursor = make_debug_cursor




from iommi.asset import Asset
from iommi.style import Style
from iommi.style_base import base

uikit = Style(
    base,
    root__assets__css=Asset.css(
        attrs=dict(
            href='https://cdn.jsdelivr.net/npm/uikit@3.6.8/dist/css/uikit.min.css',
            crossorigin='anonymous',
        ),
    ),
    Container=dict(
        tag='div',
        attrs__class={
            'uk-container': True,
        },
    ),
    Menu=dict(
        items_container=dict(
            tag='ul',
            attrs__class={'uk-nav': True, 'uk-nav-default': True},
        ),
    ),
    MenuItem=dict(
        tag='li',
        active_class_on_item=True,
        active_class='uk-active',
    ),
    Form__attrs__class={'uk-form-stacked': True},
    Field=dict(
        shortcuts=dict(
            boolean=dict(
                template='iommi/form/uikit/row_checkbox.html',
                input__attrs__class={'uk-input': False},
            ),
            radio=dict(
                input__attrs__class={'uk-input': False},
            ),
        ),
        label__attrs__class={'uk-form-label': True},
        input__attrs__class={'uk-input': True},
    ),
    Table=dict(
        attrs__class={'uk-table': True},
    ),
    Action=dict(
        shortcuts=dict(
            button__attrs__class={
                'uk-button': True,
                'uk-button-default': True,
            },
            primary__attrs__class={'uk-button-primary': True},
            delete__attrs__class={'uk-button-danger': True},
        )
    ),
    Paginator=dict(
        container__tag='ul',
        item__tag='li',
        active_item=dict(
            tag='li',
            attrs__class={'uk-active': True},
        ),
        container__attrs__class={'uk-pagination': True},
    )
)
